import pytest

from base_aux.base_lambdas.m1_lambda import *

from base_aux.valid.m1_valid_base import *
from base_aux.aux_eq.m3_eq_valid3_derivatives import *
from base_aux.base_values.m5_value_valid2_variants import *
from base_aux.base_values.m5_value_valid3_unit import *
from base_aux.buses.m1_serial1_client import *
from base_aux.buses.m1_serial2_client_derivatives import *


# =====================================================================================================================
JUST_LOAD = "JUST_LOAD"


@pytest.mark.skipif(not SerialClient.addresses_shorted__count(), reason="NO addresses_shorted__detect")
class Test__ShortedBase:
    Victim: type[SerialClient_FirstFree_Shorted]
    victim: SerialClient_FirstFree_Shorted

    SETUP_CLS__CONNECT: bool = False
    TEARDOWN__DISCONNECT: bool = True

    # -----------------------------------------------------------------------------------------------------------------
    @classmethod
    def setup_class(cls):
        pass

        class Victim(SerialClient_FirstFree_Shorted):
            RAISE_CONNECT = False

        cls.Victim = Victim
        cls.victim = cls.Victim()
        if cls.SETUP_CLS__CONNECT and not cls.victim.connect():
            msg = f"[ERROR] not found PORT shorted by Rx+Tx"
            print(msg)
            raise Exception(msg)

    @classmethod
    def teardown_class(cls):
        pass
        if hasattr(cls, "victim") and cls.victim:
            cls.victim._addresses__release()
            cls.victim.disconnect()

    def setup_method(self, method):
        pass

    def teardown_method(self, method):
        pass
        if self.TEARDOWN__DISCONNECT and hasattr(self, "victim") and self.victim:
            self.victim._addresses__release()
            self.victim.disconnect()


# =====================================================================================================================
class Test__ConnectSimple(Test__ShortedBase):
    # -----------------------------------------------------------------------------------------------------------------
    def test__address_check__resolved(self):
        self.victim = self.Victim()
        assert self.victim.address__resolve()
        assert self.victim.address_check__resolved()
        assert self.victim.address_check__occupied()
        assert self.victim.connect()


# =====================================================================================================================
class Test__AddressResolved(Test__ShortedBase):
    # -----------------------------------------------------------------------------------------------------------------
    def test__address_check__resolved(self):
        self.victim = self.Victim()
        assert self.victim.address_check__resolved() is False
        assert self.victim.connect() is True
        assert self.victim.address_check__resolved() is True
        self.victim.disconnect()
        assert self.victim.address_check__resolved() is True

        self.victim.ADDRESS = None
        assert self.victim.connect() is False
        assert self.victim.address_check__resolved() is False

        self.victim.ADDRESS = Enum__AddressAutoAcceptVariant.FIRST_FREE__SHORTED
        assert self.victim.address_check__resolved() is False
        assert self.victim.connect() is True
        assert self.victim.address_check__resolved() is True

    # -----------------------------------------------------------------------------------------------------------------
    def test__address_forget(self):
        self.victim = self.Victim()
        assert self.victim.ADDRESS == Enum__AddressAutoAcceptVariant.FIRST_FREE__SHORTED
        assert self.victim.address_check__resolved() is False
        assert self.victim.connect() is True
        assert self.victim.address_check__resolved() is True
        assert self.victim.ADDRESS != Enum__AddressAutoAcceptVariant.FIRST_FREE__SHORTED

        self.victim.address_forget()
        assert self.victim.ADDRESS == Enum__AddressAutoAcceptVariant.FIRST_FREE__SHORTED
        assert self.victim.address_check__resolved() is False
        assert self.victim.connect() is True
        assert self.victim.address_check__resolved() is True
        assert self.victim.ADDRESS != Enum__AddressAutoAcceptVariant.FIRST_FREE__SHORTED

    # -----------------------------------------------------------------------------------------------------------------
    def test__ADDRESSES(self):
        self.victim = self.Victim()
        print(self.victim.addresses_system__detect())
        assert self.victim.address_check__resolved() is False

        ports_free_started_count = self.victim.addresses_free__count()
        assert self.victim.connect() is True
        assert self.victim.connect() is True
        print(self.victim.addresses_system__detect())
        assert self.victim.address_check__resolved() is True
        assert self.victim.addresses_free__count() == ports_free_started_count - 1
        self.victim.disconnect()


# =====================================================================================================================
class Test__Connect2(Test__ShortedBase):
    # -----------------------------------------------------------------------------------------------------------------
    @classmethod
    def setup_class(cls):
        class Victim(SerialClient_FirstFree_Shorted):
            def address__validate(self):       # this is only for FIRST_FREE__ANSWER_VALID! not for FIRST_FREE__SHORTED and etc!
                return self._address__validate_shorted()

        cls.Victim = Victim
        cls.victim = cls.Victim()
        cls.victim._addresses__release()
        if not cls.victim.connect():
            msg = f"[ERROR] not found PORT shorted by Rx+Tx"
            print(msg)
            raise Exception(msg)

    def setup_method(self, method):
        self.victim.address_forget()
        self.victim.connect(_raise=False)

    # ADDRESS ---------------------------------------------------------------------------------------------------------
    def test__addresses_detect_available(self):
        assert self.Victim.addresses_system__count() > 0

    def test__address_NOTexisted(self):
        assert SerialClient.address__check_exists(address="HELLO") is False

    # AUTOCONNECT -----------------------------------------------------------------------------------------------------
    def test__ADDRESS__FIRST_VACANT(self):
        self.victim.disconnect()

        self.victim.ADDRESS = None
        assert not self.victim.connect(_raise=False)

        self.victim.ADDRESS = Enum__AddressAutoAcceptVariant.FIRST_FREE
        assert self.victim.connect(_raise=False)

        assert isinstance(self.victim.ADDRESS, str)

    def test__ADDRESS__FIRST_SHORTED(self):
        self.victim.disconnect()

        self.victim.ADDRESS = None
        assert not self.victim.connect(_raise=False)

        self.victim.ADDRESS = Enum__AddressAutoAcceptVariant.FIRST_FREE__SHORTED
        assert self.victim.connect(_raise=False)
        assert self.victim.addresses_shorted__count() > 0

    def test__ADDRESS__FIRST_ANSWER_VALID(self):
        self.victim.disconnect()

        self.victim.ADDRESS = None
        assert not self.victim.connect(_raise=False)

        self.victim.ADDRESS = Enum__AddressAutoAcceptVariant.FIRST_FREE__ANSWER_VALID
        assert self.victim.connect(_raise=False)

        # ==============
        self.victim.disconnect()
        class Victim(SerialClient):
            _ADDRESS = Enum__AddressAutoAcceptVariant.FIRST_FREE__ANSWER_VALID
            def address__validate(self) -> Union[bool, NoReturn]:
                raise Exception()

        assert not Victim().connect(_raise=False)

    # CONNECT ---------------------------------------------------------------------------------------------------------
    def test__connect_multy(self):
        assert self.victim.connect(_raise=False)
        assert self.victim.connect(_raise=False)
        assert self.victim.connect(_raise=False)

    def test__recreate_object(self):
        self.victim.disconnect()
        self.victim._address__release()

        self.victim = self.Victim()
        assert self.victim.connect(_raise=False)
        assert self.victim.write_read__last_validate(JUST_LOAD, JUST_LOAD)

        self.victim.disconnect()
        self.victim._addresses__release()

        self.victim = self.Victim()
        self.victim.connect(_raise=False)
        assert self.victim.write_read__last_validate(JUST_LOAD, JUST_LOAD)
        self.victim.disconnect()

    def test__connect_if_addr_resolved(self):
        self.victim.disconnect()

        assert self.victim.connect() is True
        assert self.victim.connect__only_if_address_resolved() is True

        self.victim.ADDRESS = None
        assert self.victim.connect(_raise=False) is False
        assert self.victim.connect__only_if_address_resolved() is None

        self.victim.ADDRESS = Enum__AddressAutoAcceptVariant.FIRST_FREE__SHORTED
        assert self.victim.connect__only_if_address_resolved() is None
        assert self.victim.connect() is True
        assert self.victim.connect__only_if_address_resolved() is True


# =====================================================================================================================
class Test__WRBase(Test__ShortedBase):
    SETUP_CLS__CONNECT: bool = True
    TEARDOWN__DISCONNECT: bool = False


# =====================================================================================================================
class Test__WR_1(Test__WRBase):
    def test__wr_single(self):
        assert self.victim.connect(_raise=False)
        assert self.victim._write("") is True
        assert self.victim.read_lines() == []

        assert self.victim._write("hello") is True
        assert self.victim.read_lines() == ["hello", ]
        assert self.victim.read_lines() == []

    def test__wr_multy(self):
        # RW single ----------------------
        for line in range(3):
            assert self.victim._write(f"hello{line}") is True

        for line in range(3):
            assert self.victim.read_line() == f"hello{line}"
        assert self.victim.read_lines() == []

        # W list ----------------------
        assert self.victim._write([f"hello{line}" for line in range(3)]) is True
        for line in range(3):
            assert self.victim.read_line() == f"hello{line}"
        assert self.victim.read_lines() == []

        # R list ----------------------
        assert self.victim._write([f"hello{line}" for line in range(3)]) is True
        assert self.victim.read_lines() == [f"hello{line}" for line in range(3)]

        assert self.victim._write([f"hello{line}" for line in range(3)]) is True
        assert self.victim.read_lines() == [f"hello{line}" for line in range(3)]

    def test__wr(self):
        assert self.victim.write_read("hello").last_output == "hello"
        assert self.victim.write_read([f"hello{line}" for line in range(3)]).list_output() == [f"hello{line}" for line in range(3)]

        # params -----------------------
        assert self.victim.write_read("hello").as_dict() == {"hello": ["hello", ], }

        assert self.victim.write_read(["11", "22"]).list_input() == ["11", "22"]
        assert self.victim.write_read(["11", "22"]).as_dict() == {"11": ["11", ], "22": ["22", ], }

        history = HistoryIO()
        history.add_io("hello", "hello")
        assert self.victim.write_read("hello").as_dict() == history.as_dict()
        assert history.check_equal_io() is True

        history = HistoryIO()
        history.add_io("11", "11")
        history.add_io("22", "22")
        assert self.victim.write_read(["11", "22"]).as_dict() == history.as_dict()
        assert history.check_equal_io() is True

    def test__wr_last(self):
        assert self.victim.write_read__last("hello") == "hello"
        assert self.victim.write_read__last(["hello1", "hello2"]) == "hello2"

    @pytest.mark.parametrize(
        argnames="write, read, _EXPECTED",
        argvalues=[
            ("hello", "hello", True),
            (["hello1", "hello2"], "hello2", True),
            (["hello1", "hello2"], "HELLO2", True),     # CASE
            (["hello1", "hello2"], "hello222", False),
            (["hello1", "hello2"], ["hello222", ], False),
            (["hello1", "hello2"], ["hello222", "hello2"], True),
            (["hello1", "hello2"], ["hello2", "hello222"], True),

            ("hello", Valid(validate_link=lambda x: x.startswith("hel")), True),
            ("hello", Valid(validate_link=lambda x: x.startswith("HEL")), False),

            ("1V", ValueUnit(unit="V"), False),
            ("1", ValueUnit(unit="V"), False),
            ("1V", ValueUnit(unit=""), True),
            ("1A", ValueUnit(unit="V"), False),

            ("3", ValueVariants(variants=[3, ]), True),
        ],
    )
    def test__wr_last_validate(self, write, read, _EXPECTED):
        func_link = lambda: self.victim.write_read__last_validate(write, read)
        Lambda(func_link).check_expected__assert(_EXPECTED)

    @pytest.mark.parametrize(
        argnames="write, read, _EXPECTED",
        argvalues=[
            ("hello", "he.*", True),
            ("hello", ["he.*", ], True),
            ("hello", ["HE.*", ], True),
            ("hello", ["hello123", ], False),
            ("hello", ["hello123", "HELLO"], True),
        ],
    )
    def test__wr_last_validate_regexp(self, write, read, _EXPECTED):
        func_link = lambda: self.victim.write_read__last_validate_regexp(write, read)
        Lambda(func_link).check_expected__assert(_EXPECTED)

    def test__wr_ReadFailPattern(self):
        self.victim.RAISE_READ_FAIL_PATTERN = True
        try:
            self.victim.write_read("123 FAil 123")
        except:
            assert True
        else:
            assert False

        self.victim.RAISE_READ_FAIL_PATTERN = False
        assert self.victim.write_read("123 FAil 123").last_output == "123 FAil 123"

    def test__r_all(self):
        assert self.victim._write([f"hello{i}" for i in range(3)]) is True
        assert self.victim.read_lines() == [f"hello{i}" for i in range(3)]

    def test__write_args_kwargs(self):
        assert self.victim.write_read("hello").last_output == "hello"
        assert self.victim.write_read("hello", args=[1, 2]).last_output == "hello 1 2"
        assert self.victim.write_read("hello", kwargs={"CH1": 1}).last_output == "hello CH1=1"
        assert self.victim.write_read("hello", args=[1, 2], kwargs={"CH1": 1}).last_output == "hello 1 2 CH1=1"

    def test__CMD_PREFIX(self):
        self.victim.PREFIX = "DEV:01:"
        assert self.victim.write_read("hello").last_output == f"{self.victim.PREFIX}hello"
        assert self.victim.write_read("hello 12").last_output == f"{self.victim.PREFIX}hello 12"

        self.victim.PREFIX = ""
        assert self.victim.write_read("hello").last_output == "hello"


# =====================================================================================================================
class Test__WR_2_Getattr(Test__WRBase):
    def test__getattr__1_STARTSWITH(self):
        assert self.victim.send__hello() == "hello"

    def test__getattr__2_cmd_direct(self):
        assert self.victim.hello() == "hello"
        assert self.victim.hello(12) == "hello 12"
        assert self.victim.hello(12, 13) == "hello 12 13"
        assert self.victim.hello("12 13") == "hello 12 13"
        assert self.victim.hello(CH1=12, CH2=13) == "hello CH1=12 CH2=13"
        assert self.victim.hello("?") == "hello ?"

        assert self.victim.hello(12, kwarg1=1) == "hello 12 kwarg1=1"
        assert self.victim.send__hello(12, kwarg1=1) == "hello 12 kwarg1=1"

    def test__getattr__3_cmd_w_args(self):
        assert self.victim.hello__11() == "hello 11"
        assert self.victim.hello__11__22() == "hello 11 22"
        assert self.victim.hello__11__22(33) == "hello 11 22 33"
        assert self.victim.hello__11__22(33, 44, kwarg1=1, kwarg2=2) == "hello 11 22 33 44 kwarg1=1 kwarg2=2"

        assert self.victim.send__hello__11__22(33, 44, kwarg1=1, kwarg2=2) == "hello 11 22 33 44 kwarg1=1 kwarg2=2"

    def test__getattr__0_PREFIX(self):
        self.victim.PREFIX = "DEV:01:"
        assert self.victim.hello() == f"{self.victim.PREFIX}hello"
        assert self.victim.hello(12) == f"{self.victim.PREFIX}hello 12"
        self.victim.PREFIX = ""
        assert self.victim.hello() == f"hello"


# =====================================================================================================================
class Test__WR_3_ValueUnit(Test__WRBase):
    def test__1(self):
        assert self.victim.send__123() == "123"
        # NOTE: use other EqValid_ParseSingleNum* - NO!!!
        assert self.victim.send__123() == 123
        assert self.victim.send__123() > 122
        assert self.victim.send__123() < 123.1

        assert self.victim.send__123v() > 122
        assert self.victim.send__123v() + 1 == 124


# =====================================================================================================================



import pytest

from base_aux.buses.m1_serial2_client_derivatives import *


# =====================================================================================================================
# @pytest.mark.skip
class Test__Shorted_validateModel_InfinitRW:
    """
    VALIDATE ADAPTERS fro DECODУ ERRORS
    """
    Victim: type[SerialClient_FirstFree_Shorted] = SerialClient_FirstFree_Shorted
    victim: SerialClient_FirstFree_Shorted

    # @classmethod
    # def setup_class(cls):
    #     pass
    #
    # @classmethod
    # def teardown_class(cls):
    #     if cls.victim:
    #         cls.victim.disconnect()
    #
    # def setup_method(self, method):
    #     pass

    def teardown_method(self, method):
        pass
        if self.victim:
            self.victim.disconnect()

    # -----------------------------------------------------------------------------------------------------------------
    @pytest.mark.skip
    def test__shorted(self):
        """
        connect shorted port and start this code!
        wait for first error and see the step number
        write as result in SerialClient docstring
        """
        # PREPARE ------------------------
        self.Victim.BAUDRATE = 115200
        self.victim = self.Victim()
        self.victim.test__shorted()

    @pytest.mark.skip
    def test__real_device(self):    # TODO: fix it!!!
        """
        in case of validate real device like ATC/PTB
        DONT FORGET TO CHANGE address__validate for your device!!!
        """
        # PREPARE ------------------------
        VALIDATION_CMD, VALIDATION_ANSWER = ("GET NAME", "ATC 03")

        class Victim(SerialClient_FirstFree_Shorted):
            def address__validate(self) -> bool:
                return self.write_read__last_validate(*VALIDATION_CMD)

        self.victim = Victim()
        if not self.victim.connect():
            msg = f"[ERROR] not found PORT validated"
            print(msg)
            raise Exception(msg)

        # START WORKING ------------------
        index = 0
        while True:
            index += 1
            load = f"step{index}"
            print(load)
            assert self.victim.address__validate() is True


# =====================================================================================================================
if __name__ == "__main__":
    Test__Shorted_validateModel_InfinitRW().test__shorted()


# =====================================================================================================================



# from typing import *
#
# from base_aux.aux_text.m5_re1_rexp import *
# from base_aux.aux_iter.m1_iter_aux import *
# from base_aux.aux_attr.m1_attr2_nest8_iter_name_value import *
# from base_aux.base_nest_dunders.m3_nest_init_annots_attrs_by_kwargs import *
# from base_aux.aux_datetime.m1_datetime import *
#
#
# # =====================================================================================================================
# class Base_MsgLinesNamed:
#     """
#     GOAL
#     ----
#     use structured object for msg
#     with simply replacing special blocks
#     instead of manually parsing
#
#     SPECIALLY CREATED FOR
#     ---------------------
#     Alerts Telegram
#     """
#     _sep = "-"*20
#
#     ALERT: str
#     _sep1: str = _sep
#     BODY: str
#     _sep2: str = _sep
#     TS: TimeStampRenewStr = TimeStampRenewStr()
#
#     def __init__(self, text: str):
#         pass
#
#     def __str__(self):
#         result = ""
#         for name, value in self.iter_exisded_name_value():
#             name: str
#             if name.startswith("_"):
#                 result += f"[{name}]{value}"
#             else:
#                 result += f"{value}"
#
#     def __repr__(self):
#         return str(self)
#
#     def iter_exisded_name_value(self) -> tuple[str, Any]:
#         for name in self.__annotations__:
#             try:
#                 yield name, getattr(self, name)
#             except:
#                 pass
#
#
# # =====================================================================================================================



import pytest
from typing import *
from base_aux.servers.m0_url import Url
from base_aux.base_lambdas.m1_lambda import *


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="kwargs_source, kwargs_resolve, _EXPECTED",
    argvalues=[
        (dict(), dict(), "http://localhost:80/"),

        (dict(host="host"), dict(), "http://host:80/"),

        (dict(host="host", route="route"), dict(), "http://host:80/route"),

        (dict(host="host", route="route"), dict(host="host2"), "http://host2:80/route"),

        (dict(host="host", route="route"), dict(port=8080), "http://host:8080/route"),

        # ROUTE ---
        (dict(host="host", route="route"), dict(route=""), "http://host:80/"),
        (dict(host="host", route="route"), dict(route="route2"), "http://host:80/route2"),
        (dict(host="host", route="route"), dict(route="/route2"), "http://host:80/route2"),
        (dict(host="host", route="route"), dict(route="//route2"), "http://host:80/route2"),
        (dict(host="host", route="route"), dict(route="///route2"), "http://host:80/route2"),

        (dict(host="host", route="route"), dict(route="route2/"), "http://host:80/route2/"),
        (dict(host="host", route="route"), dict(route="route2///"), "http://host:80/route2///"),

        (dict(host="host", route="route"), dict(route="route2/3"), "http://host:80/route2/3"),
        (dict(host="host", route="route"), dict(route="2/3"), "http://host:80/2/3"),
        (dict(host="host", route="route"), dict(route=(2, 3)), "http://host:80/2/3"),
        (dict(host="host", route="route"), dict(route=[2, 3]), "http://host:80/2/3"),
        (dict(host="host", route="route"), dict(route=[]), "http://host:80/"),

    ]
)
def test__url(kwargs_source, kwargs_resolve, _EXPECTED):
    Lambda(Url(**kwargs_source).resolve, **kwargs_resolve).check_expected__assert(_EXPECTED)


# =====================================================================================================================



import time


# =====================================================================================================================
class Stopwatch:
    """
    GOAL
    ----
    create (at last) timer to check time passed
    """
    # TODO: FINISH! add tests
    # TODO: FINISH! add tests
    # TODO: FINISH! add tests
    # TODO: FINISH! add tests
    # TODO: FINISH! add tests
    # TODO: FINISH! add tests
    # TODO: FINISH! add tests
    # TODO: FINISH! add tests
    # TODO: FINISH! add tests

    START_ON_INIT: bool = False

    def __init__(self, start: bool = None):
        self.time_initial: float = time.time()  # TODO: deprecate? use only one?
        self.time_started: float | None = None
        self.time_finished: float | None = None

        self._paused__state: bool = False
        self._paused__last_ts: float | None = None
        self._paused__summary: float = 0

        if start is not None:
            if start:
                self.time_started: float = self.time_initial
        elif self.START_ON_INIT:
            self.time_started: float = self.time_initial

    # -----------------------------------------------------------------------------------------------------------------
    def _clear(self) -> None:
        self.time_started = None
        self._paused__state = False
        self._paused__last_ts = None
        self._paused__summary = 0

    # -----------------------------------------------------------------------------------------------------------------
    def get_elapsed_time__from_start(self) -> float:
        """
        GOAL
        ----
        with/including paused time!
        """
        try:
            return time.time() - self.time_started
        except:
            return 0

    def get_execution_time__from_start(self) -> float:
        """
        GOAL
        ----
        without/not including paused time!
        """
        raise NotImplementedError()

    # -----------------------------------------------------------------------------------------------------------------
    def start(self) -> None:
        raise NotImplementedError()

    def stop(self) -> None:
        raise NotImplementedError()

    # -----------------------------------------------------------------------------------------------------------------
    def pause(self) -> None:
        raise NotImplementedError()

    def resume(self) -> None:
        raise NotImplementedError()



    # -----------------------------------------------------------------------------------------------------------------
    def wait_execution__from_start(self, target: float) -> None:
        raise NotImplementedError()

    async def aio_wait_execution__from_start(self, target: float) -> None:
        raise NotImplementedError()


# =====================================================================================================================
class StopwatchStarted(Stopwatch):
    START_ON_INIT = True


# =====================================================================================================================



from typing import Any, Self

from base_aux.base_monkeys.m0_base import Mark_Monkey
from base_aux.base_nest_dunders.m4_ga_self import NestGa_Self
from base_aux.base_nest_dunders.m2_str1_cls import *


# =====================================================================================================================
class Monkey_GaSelf(NestGa_Self, Mark_Monkey):
    """
    GOAL
    ----
    just show that final class is exactly the monkey - no more differences from NestGa_Self
    """


# =====================================================================================================================
class Monkey_GaSelf_CallResult(NestGa_Self, Mark_Monkey):
    """
    GOAL
    ----
    in test suits when we need some object like
        victim.any.attr.chain.could.be.here()

    EXAMPLE
    -------
    use any attr chain
        victim = Monkey_GaSelf_CallResult(5)
        assert victim.hello == victim
        assert victim.hello() == 5

        assert victim.hello.world == victim
        assert victim.hello.world() == 5
    """
    def __init__(self, call_result: Any = None) -> None:
        self._call_result: Any = call_result

    def __call__(self) -> Any:
        return self._call_result

    def __eq__(self, other: Self | Any) -> bool:
        if isinstance(other, self.__class__):
            return self._call_result == other._call_result
        else:
            return False

    # -----------------------------------------------------------------------------------------------------------------
    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self._call_result=})"

    def __repr__(self) -> str:
        return str(self)


# =====================================================================================================================



import re

from base_aux.aux_attr.m4_dump0_dumped import *
from base_aux.base_lambdas.m1_lambda import *
# from base_aux.aux_iter.m1_iter_aux import *   # dont add! import error!
from base_aux.aux_eq.m3_eq_valid1_base import Base_EqValid
from base_aux.base_types.m0_static_typing import TYPING


# =====================================================================================================================
class Base_AttrAux(NestInit_Source):
    """
    NOTE
    ----
    1. if there are several same aux_attr in different cases - you should resolve it by yourself!- what it is about???!
    2. ITERAING names
        - DONT intend iterate PRIVATEs! - it is just for information! yoг cant use it to get values!

    ANNOTS
    ------
    names intended in annots

    DIFFERENCE from AnnotAux
    ------------------------
    1/ iter names
        - dump dict
        - cmp eq by dumped dict
    2/ AnnotAux is not listing methods! list only annots!

    FIXME: add skip methods??? seems it need!
    """
    # SOURCE: Any
    SOURCE: Any = AttrDumped
    SKIP_NAMES: tuple[str | Base_EqValid, ...] = ()

    _ATTRS_STYLE: EnumAdj_AttrAnnotsOrExisted = EnumAdj_AttrAnnotsOrExisted.ATTRS_EXISTED
    _ANNOTS_DEPTH: EnumAdj_AnnotsDepthAllOrLast = EnumAdj_AnnotsDepthAllOrLast.ALL_NESTED

    # =================================================================================================================
    def __init__(self, source: Any = NoValue, *skip_names: str | Base_EqValid) -> None:
        super().__init__(source)
        if skip_names:
            self.SKIP_NAMES = skip_names

    # =================================================================================================================
    def name__check_is_method(self, name_original: str) -> bool:
        try:
            value = getattr(self.SOURCE, name_original)
        except:
            return False

        return TypeAux(value).check__callable_meth()

    def name__check_is_private(self, name: str) -> bool:
        name = str(name)

        if re.fullmatch(r"_.+__.+", name) is not None or name.startswith("__"):
            return True

        return False

    # =================================================================================================================
    def ITER_NAMES_BY_STYLE(self) -> Iterable[TYPING.ATTR_FINAL]:
        """
        GOAL
        ----
        iterator names depends on
            - existed
            - annots
        """
        if self._ATTRS_STYLE == EnumAdj_AttrAnnotsOrExisted.ATTRS_EXISTED:
            yield from self.iter__dirnames_original_not_builtin()
        elif self._ATTRS_STYLE == EnumAdj_AttrAnnotsOrExisted.ANNOTS_ONLY:
            yield from self.iter__annot_names()
        else:
            raise Exc__Incompatible_Data(f"{self._ATTRS_STYLE=}/{self._ANNOTS_DEPTH=}")

    # -----------------------------------------------------------------------------------------------------------------
    pass

    # def __contains__(self, item: str):      # IN=DONT USE IT! USE DIRECT METHOD anycase__check_exists
    #     return self.anycase__check_exists(item)

    def iter__dirnames_and_annots(self) -> Iterable[TYPING.ATTR_FINAL]:
        """
        GOAL
        ----
        when you try to set_ic you must intend names from ANNOTS! otherwise you can set

        SPECIALLY CREATED FOR
        ---------------------
        sai_ic only!
        """
        yield from self.iter__dirnames_original_not_builtin()
        yield from self.iter__annot_names()

    def iter__dirnames_original_not_builtin(self) -> Iterable[TYPING.ATTR_FINAL]:
        """
        GOAL
        ----
        BASE ITERATOR for all dirnames codeOriginal userNames!!!

        1/ iter ALL (without builtins)!!! - so it makes BASE iterator for overall names
        2/ use private names in ORIGIN!
        3/ if you need apply any filter on
        """
        for name in dir(self.SOURCE):
            # filter-1 skip ----------
            if name in self.SKIP_NAMES:
                continue

            # filter-2 builtin ----------
            if name.startswith("__") and name.endswith("__"):
                continue

            # rename private original ----------
            name = self.try_rename__private_original(name)

            # if self.name__check_is_method(name):    # FIXME: DONT USE HERE!!! or resolve what to do!!!
            #     continue

            # direct user attr ----------
            # print(f"{name=}")
            yield name

        yield from []

    # =================================================================================================================
    def iter__names_filter(self, attr_level: EnumAdj_AttrScope = EnumAdj_AttrScope.NOT_PRIVATE) -> Iterable[TYPING.ATTR_FINAL]:
        """
        GOAL
        ----
        iter names with filter
        from ITER_NAMES_EXISTED_OR_ANNOTS
        """
        # -------------------------------------------------
        for name in self.ITER_NAMES_BY_STYLE():
            if name in self.SKIP_NAMES:
                continue

            if attr_level == EnumAdj_AttrScope.NOT_PRIVATE:
                if not self.name__check_is_private(name):
                    yield name

            elif attr_level == EnumAdj_AttrScope.NOT_HIDDEN:
                if not name.startswith("_"):
                    yield name

            elif attr_level == EnumAdj_AttrScope.PRIVATE:
                if self.name__check_is_private(name):
                    name = self.try_rename__private_original(name)
                    yield name

            elif attr_level == EnumAdj_AttrScope.ALL:
                yield name

            else:
                raise Exc__Incompatible_Data(f"{attr_level=}")

    def iter__names_filter__not_hidden(self) -> Iterable[TYPING.ATTR_FINAL]:
        """
        NOTE
        ----
        hidden names are more simple to detect then private!
        cause of private methods(!) changes to "_<ClsName><__MethName>"
        """
        return self.iter__names_filter(EnumAdj_AttrScope.NOT_HIDDEN)

    def iter__names_filter__not_private(self) -> Iterable[TYPING.ATTR_FINAL]:
        """
        NOTE
        ----
        BEST WAY TO USE EXACTLY iter__not_private
        """
        return self.iter__names_filter(EnumAdj_AttrScope.NOT_PRIVATE)

    def iter__names_filter__private(self) -> Iterable[TYPING.ATTR_FINAL]:
        """
        NOTE
        ----
        BUILTIN - NOT INCLUDED!

        GOAL
        ----
        collect all privates in original names! without ClassName-Prefix
        """
        return self.iter__names_filter(EnumAdj_AttrScope.PRIVATE)

    # def __iter__(self):     # DONT USE IT! USE DIRECT METHODS
    #     yield from self.iter__not_hidden()

    # -----------------------------------------------------------------------------------------------------------------
    def _iter_mro(self) -> Iterable[type]:
        """
        GOAL
        ----
        iter only important user classes from mro
        """
        yield from TypeAux(self.SOURCE).iter_mro_user(
            # NestGAI_AnnotAttrIC,
            # NestGSAI_AttrAnycase,
            # NestGA_AnnotAttrIC, NestGI_AnnotAttrIC,
            # NestSA_AttrAnycase, NestSI_AttrAnycase,
        )

    # -----------------------------------------------------------------------------------------------------------------
    def iter__annot_names(self) -> Iterable[TYPING.ATTR_FINAL]:
        """
        iter all (with not existed)
        """
        yield from self.dump_dict__annot_types()

    def iter__annot_values(self) -> Iterable[Any]:
        """
        NOTE
        ----
        only existed and not Raised! - skip otherwise!
        """
        # TODO: decide what to do with 1. notExisted value, 2. existedRaised (property) 3. existedCallable
        for name in self.iter__annot_names():
            try:
                yield self.gai_ic(name)
            except:
                pass

    # -----------------------------------------------------------------------------------------------------------------
    def list__annots(self) -> list[TYPING.ATTR_FINAL]:
        """
        GOAL
        ----
        symply just add list instead of iterator!
        """
        return list(self.dump_dict__annot_types())

    # =================================================================================================================
    def reinit__mutable_cls_values(self) -> None:
        """
        GOAL
        ----
        reinit default mutable values from class dicts/lists on instantiation.
        usually intended blank values.

        REASON
        ------
        for dataclasses you should use field(dict) but i think it is complicated (but of cause more clear)

        SPECIALLY CREATED FOR
        ---------------------
        Base_AttrKit
        """
        try:
            source_is_cls = issubclass(self.SOURCE, object)
        except:
            source_is_cls = False

        for attr in self.iter__names_filter__not_private():
            if source_is_cls:
                try:
                    value = getattr(self.SOURCE, attr)
                except:
                    continue
            else:
                try:
                    value_cls = getattr(self.SOURCE.__class__, attr)
                    value = getattr(self.SOURCE, attr)  # important using instanceValue!!!
                except:
                    continue
                if value is not value_cls:
                    # expecting that if you pass smth that it is real exact final value! which you dont need to make a copy!
                    continue

            if isinstance(value, dict):
                setattr(self.SOURCE, attr, dict(value))
            elif isinstance(value, list):
                setattr(self.SOURCE, attr, list(value))
            elif isinstance(value, set):
                setattr(self.SOURCE, attr, set(value))

    def reinit__annots_by_None(self) -> None:
        """
        GOAL
        ----
        set None for all annotated aux_attr! even not existed!
        """
        for name in self.iter__annot_names():
            self.sai_ic(name, None)

    def reinit__annots_by_types(self, not_existed: bool = None) -> None:
        """
        GOAL
        ----
        delattr all annotated aux_attr!
        """
        for name, value in self.dump_dict__annot_types().items():
            if not_existed and hasattr(self.SOURCE, name):
                continue

            value = TypeAux(value).type__init_value__default()
            self.sai_ic(name, value)

    # =================================================================================================================
    def try_rename__private_original(self, dirname: str) -> TYPING.ATTR_FINAL:
        """
        GOAL
        ----
        try rename attr name for private code-original or return old name
        using name (from dir(obj)) return user-friendly code-original name!

        REASON
        ------
        here in example - "__hello" will never appear directly!!!
        class Cls:
            ATTR1 = 1
            def __hello(self, *args) -> None:
                kwargs = dict.fromkeys(args)
                self.__init_kwargs(**kwargs)

        name='_Cls__hello' hasattr(self.SOURCE, name)=True
        name='__class__' hasattr(self.SOURCE, name)=True
        name='__delattr__' hasattr(self.SOURCE, name)=True
        name='__dict__' hasattr(self.SOURCE, name)=True
        name='__dir__' hasattr(self.SOURCE, name)=True
        name='__doc__' hasattr(self.SOURCE, name)=True
        ///
        name='ATTR1' hasattr(self.SOURCE, name)=True
        """
        # filter not hidden -------
        if not dirname.startswith("_"):
            return dirname

        # filter private builtin -------
        if dirname.startswith("__"):
            return dirname

        # parse private code-original name -------
        if re.fullmatch(r"_.+__.+", dirname):
            # print(f"{dirname=}")
            # print(f"{self.SOURCE=}")
            try:
                mro = self.SOURCE.__mro__   # instance has no attr __mro__ - RAISE
                # print(f"11111{mro=}")
                if not isinstance(mro, tuple):
                    # BUT it could be instance ... with GETATTR!!! then
                    mro = self.SOURCE.__class__.__mro__
                    # print(f"33333{mro=}")

            except:
                # if SOURCE was instance (NotClass) - direct check for isClass - is not correct!
                mro = self.SOURCE.__class__.__mro__
                # print(f"22222{mro=}")

            for cls in mro:
                if dirname.startswith(f"_{cls.__name__}__"):
                    name_original = dirname.replace(f"_{cls.__name__}", "")
                    return name_original

        return dirname

    # -----------------------------------------------------------------------------------------------------------------
    def name_ic__get_original(self, name_index: TYPING.ATTR_DRAFT) -> TYPING.ATTR_FINAL | None:
        """
        get attr name_index in original register
        """
        # print(f"{name_index=}")
        name_index = str(name_index)
        name_index = str(name_index).strip()

        # name as index for annots -------
        index = None
        try:
            index = int(name_index)
        except:
            pass

        if isinstance(index, int) and index is not None:
            return self.list__annots()[index]  # dont place in try sentence

        # name as str for annots/attrs ------
        if not name_index:
            return

        for name_original in [*self.iter__dirnames_original_not_builtin(), ]:  # *self.list__annots()], - make infinitive execution!
            if name_original.lower() == name_index.lower():
                return name_original

        return

    def name_ic__check_exists(self, name_index: TYPING.ATTR_DRAFT) -> bool:
        return self.name_ic__get_original(name_index) is not None

    def name__check_have_value(self, name_index: TYPING.ATTR_DRAFT) -> bool:
        """
        GOAL
        ----
        check attr really existed!
        separate exc on getattr (like for property) and name-not-existed.
        used only due to annots!

        SPECIALLY CREATED FOR
        ---------------------
        dump_dict - because in there if not value exists - logic is differ from base logic! (here we need to pass!)
        """
        name_final = self.name_ic__get_original(name_index)
        if name_final:
            return hasattr(self.SOURCE, name_final)
        else:
            return False

    # -----------------------------------------------------------------------------------------------------------------
    def annots__ensure(self) -> None:
        """
        GOAL
        ----
        unsure access to __annotations__ if it was not created on class!

        REASON
        ------
        if class have not any annotations and you will access them over instance - raise!
        but if you will first access annotations over class - no raise!

            AttributeError: 'AttrDumped' object has no attribute '__annotations__'

        SPECIALLY CREATED FOR
        ---------------------
        annots__append
        """
        try:
            self.SOURCE.__class__.__annotations__
        except:
            pass

        try:
            self.SOURCE.__annotations__
        except:
            pass

    def annots__append_with_values(self, **kwargs: type | Any) -> AttrDumped | Any:
        """
        GOAL
        ----
        append new annot in last position

        BEST USAGE
        ----------
        create NEW OBJECT

        SPECIALLY CREATED FOR
        ---------------------
        TextFormatted
        """
        self.annots__ensure()
        annots: dict[str, type] = self.SOURCE.__annotations__
        for key, value in kwargs.items():
            if value is None:
                value_type = Any
            elif not TypeAux(value).check__class():
                value_type = type(value)
            else:
                value_type = value

            # set type
            if key.lower() not in [key_orig.lower() for key_orig in annots]:
                annots.update({key: value_type})

            # set value
            if value != value_type:
                self.sai_ic(key, value)

        return self.SOURCE

    def annots__get_not_defined(self) -> list[TYPING.ATTR_FINAL]:
        """
        GOAL
        ----
        return list of not defined annotations

        SPECIALLY CREATED FOR
        ---------------------
        annot__check_all_defined
        """
        result = []
        nested = self.dump_dict__annot_types()
        for key in nested:
            if not self.name_ic__check_exists(key):
                result.append(key)
        return result

    def annots__check_all_defined(self) -> bool:
        """
        GOAL
        ----
        check if all annotated aux_attr have value!
        """
        return not self.annots__get_not_defined()

    def annots__check_all_defined_or_raise(self) -> bool | NoReturn:
        """
        GOAL
        ----
        check if all annotated aux_attr have value!
        """
        not_defined = self.annots__get_not_defined()
        if not_defined:
            dict_type = self.dump_dict__annot_types()
            msg = f"{not_defined=} in {dict_type=}"
            raise Exc__NotExistsNotFoundNotCreated(msg)

        return True

    # =================================================================================================================
    def gai_ic(self, name_index: TYPING.ATTR_DRAFT) -> Any | Callable | NoReturn:
        """
        GOAL
        ----
        get attr value by name_index in any register
        no execution/resolving! return pure value as represented in object!
        """
        name_original = self.name_ic__get_original(name_index)

        if name_index == "__name__":        # this is a crutch! костыль!!!!
            result = "ATTRS"
            result = self.SOURCE.__class__.__name__
            return result

        if name_original is None:
            raise IndexError(f"{name_index=}/{self=}")

        return getattr(self.SOURCE, name_original)

    def sai_ic(self, name_index: TYPING.ATTR_DRAFT, value: Any) -> None | NoReturn:
        """
        get attr value by name_index in any register
        no execution! return pure value as represented in object!

        NoReturn - in case of not accepted names when setattr
        """
        name_original: str = self.name_ic__get_original(name_index)
        if name_original is None:
            name_original = name_index

        if not name_original:
            raise IndexError(f"{name_index=}/{self=}")

        # NOTE: you still have no exc with setattr(self.SOURCE, "    HELLO", value) and ""
        setattr(self.SOURCE, name_original, value)
        pass

    def dai_ic(self, name_index: TYPING.ATTR_DRAFT) -> None:
        name_original = self.name_ic__get_original(name_index)
        if name_original is None:
            return      # already not exists

        delattr(self.SOURCE, name_original)

    # -----------------------------------------------------------------------------------------------------------------
    def gai_ic__callable_resolve(self, name_index: TYPING.ATTR_DRAFT, callables_resolve: EnumAdj_CallResolveStyle = EnumAdj_CallResolveStyle.DIRECT) -> Any | Callable | EnumAdj_CallResolveStyle | NoReturn:
        """
        SAME AS
        -------
        CallableAux(*).resolve_*

        WHY NOT-1=CallableAux(*).resolve_*
        ----------------------------------
        it is really the same, BUT
        1. additional try for properties (could be raised without calling)
        2. cant use here cause of Circular import accused
        """
        # resolve property --------------
        # result_property = CallableAux(getattr).resolve(callables_resolve, self.SOURCE, realname)
        # TypeAux

        try:
            value = self.gai_ic(name_index)
        except Exception as exc:
            if callables_resolve == EnumAdj_CallResolveStyle.SKIP_RAISED:
                return VALUE_SPECIAL.SKIPPED
            elif callables_resolve == EnumAdj_CallResolveStyle.EXC:
                return exc
            elif callables_resolve == EnumAdj_CallResolveStyle.RAISE_AS_NONE:
                return None
            elif callables_resolve == EnumAdj_CallResolveStyle.RAISE:
                raise exc
            elif callables_resolve == EnumAdj_CallResolveStyle.BOOL:
                return False
            else:
                raise exc

        # resolve callables ------------------
        result = Lambda(value).resolve__style(callables_resolve)
        return result

    # -----------------------------------------------------------------------------------------------------------------
    def sai__by_args_kwargs(self, *args: Any, **kwargs: dict[str, Any]) -> Any | NoReturn:
        """
        MAIN ITEA
        ----------
        LOAD MEANS basically setup final values for final not callables values!
        but you can use any types for your own!
        expected you know what you do and do exactly ready to use final values/not callables in otherObj!
        """
        self.sai__by_args(*args)
        self.sai__by_kwargs(**kwargs)

        return self.SOURCE

    def sai__by_args(self, *args: Any) -> Any | NoReturn:
        for index, value in enumerate(args):
            self.sai_ic(index, value)

        return self.SOURCE

    def sai__by_kwargs(self, **kwargs: dict[str, Any]) -> Any:
        for name, value in kwargs.items():
            self.sai_ic(name, value)

        return self.SOURCE

    # =================================================================================================================
    def dump_dict__annot_types(self) -> dict[str, type[Any]]:
        """
        GOAL
        ----
        get all annotations in correct order (nesting available)!

        RETURN
        ------
        keys - all attr names (defined and not)
        values - Types!!! not instances!!!
        """
        result = {}
        if self._ANNOTS_DEPTH == EnumAdj_AnnotsDepthAllOrLast.LAST_CHILD:
            result = dict(self.SOURCE.__annotations__)

        elif self._ANNOTS_DEPTH == EnumAdj_AnnotsDepthAllOrLast.ALL_NESTED or True:
            for cls in self._iter_mro():
                try:
                    _result_i = dict(cls.__annotations__)
                except:
                    break

                _result_i.update(result)
                result = _result_i


        # else:
        #     raise Exc__Incompatible_Data(f"{self._ANNOTS_DEPTH=}")

        # rename private original -------------------
        result_final: dict[str, type[Any]] = dict()
        for name, value in result.items():
            # filter private external ----------
            name = self.try_rename__private_original(name)
            result_final.update({name: value})
        return result_final

    # -----------------------------------------------------------------------------------------------------------------
    def dump_dict(
            self,
            *skip_names: str | Base_EqValid,
            callables_resolve: EnumAdj_CallResolveStyle = EnumAdj_CallResolveStyle.EXC,
    ) -> dict[str, Any | Callable | Exception] | NoReturn:
        """
        MAIN IDEA
        ----------
        BUMPS MEANS basically save final values for all (even any dynamic/callables) values! or only not callables!
        SKIP NOT EXISTED ANNOTS!!!

        NOTE
        ----
        DUMP WITHOUT PRIVATE NAMES

        GOAL
        ----
        make a dict from any object from aux_attr (not hidden)

        SPECIALLY CREATED FOR
        ---------------------
        using any object as rules for Translator
        """
        skip_names = skip_names or []
        result = {}
        for name in self.iter__names_filter__not_private():
            # skip is attr not exist
            if not self.name__check_have_value(name):
                continue
            if name in skip_names:
                continue

            value = self.gai_ic__callable_resolve(name_index=name, callables_resolve=callables_resolve)
            if value is VALUE_SPECIAL.SKIPPED:
                continue
            result.update({name: value})

        return result

    def dump_dict__resolve_exc(self, *skip_names: str | Base_EqValid) -> dict[str, Any | Exception]:
        """
        MAIN DERIVATIVE!
        """
        return self.dump_dict(*skip_names, callables_resolve=EnumAdj_CallResolveStyle.EXC)

    def dump_dict__direct(self, *skip_names: str | Base_EqValid) -> TYPING.KWARGS_FINAL:
        return self.dump_dict(*skip_names, callables_resolve=EnumAdj_CallResolveStyle.DIRECT)

    def dump_dict__skip_callables(self, *skip_names: str | Base_EqValid) -> TYPING.KWARGS_FINAL:
        return self.dump_dict(*skip_names, callables_resolve=EnumAdj_CallResolveStyle.SKIP_CALLABLE)

    def dump_dict__skip_raised(self, *skip_names: str | Base_EqValid) -> dict[str, Any] | NoReturn:
        return self.dump_dict(*skip_names, callables_resolve=EnumAdj_CallResolveStyle.RAISE)

    # -----------------------------------------------------------------------------------------------------------------
    def dump_obj(
            self,
            *skip_names: str | Base_EqValid,
            callables_resolve: EnumAdj_CallResolveStyle = EnumAdj_CallResolveStyle.EXC,
    ) -> AttrDumped | NoReturn:
        data = self.dump_dict(*skip_names, callables_resolve=callables_resolve)
        obj = AttrAux_Existed(AttrDumped()).sai__by_args_kwargs(**data)
        return obj

    def dump_obj__resolve_exc(self, *skip_names: str | Base_EqValid) -> dict[str, Any | Exception]:
        """
        MAIN DERIVATIVE!
        """
        return self.dump_obj(*skip_names, callables_resolve=EnumAdj_CallResolveStyle.EXC)

    def dump_obj__direct(self, *skip_names: str | Base_EqValid) -> TYPING.KWARGS_FINAL:
        return self.dump_obj(*skip_names, callables_resolve=EnumAdj_CallResolveStyle.DIRECT)

    def dump_obj__skip_callables(self, *skip_names: str | Base_EqValid) -> TYPING.KWARGS_FINAL:
        return self.dump_obj(*skip_names, callables_resolve=EnumAdj_CallResolveStyle.SKIP_CALLABLE)

    def dump_obj__skip_raised(self, *skip_names: str | Base_EqValid) -> dict[str, Any] | NoReturn:
        return self.dump_obj(*skip_names, callables_resolve=EnumAdj_CallResolveStyle.RAISE)

    # -----------------------------------------------------------------------------------------------------------------
    def dump_str__pretty(
            self,
            *skip_names: str | Base_EqValid,
            callables_resolve: EnumAdj_CallResolveStyle = EnumAdj_CallResolveStyle.EXC,
    ) -> str:
        try:
            result = f"{self.SOURCE.__class__.__name__}(Attributes):"
        except:
            result = f"{self.SOURCE.__name__}(Attributes):"

        for key, value in self.dump_dict(*skip_names, callables_resolve=callables_resolve).items():
            result += f"\n    {key}={value}"
        else:
            result += f"\nEmpty=Empty"

        return result


# =====================================================================================================================
@final
class AttrAux_Existed(Base_AttrAux):
    _ATTRS_STYLE: EnumAdj_AttrAnnotsOrExisted = EnumAdj_AttrAnnotsOrExisted.ATTRS_EXISTED
    _ANNOTS_DEPTH: EnumAdj_AnnotsDepthAllOrLast = EnumAdj_AnnotsDepthAllOrLast.ALL_NESTED


# ---------------------------------------------------------------------------------------------------------------------
@final
class AttrAux_AnnotsAll(Base_AttrAux):
    """
    GOAL
    ----
    work with all __annotations__
        from all nested classes
        in correct order

    RULES
    -----
    1. nesting available with correct order!
        class ClsFirst(BreederStrStack):
            atr1: int
            atr3: int = None

        class ClsLast(BreederStrStack):
            atr2: int = None
            atr4: int

        for key, value in ClsLast.annotations__get_nested().items():
            print(f"{key}:{value}")

        # atr1:<class 'int'>
        # atr3:<class 'int'>
        # atr2:<class 'int'>
        # atr4:<class 'int'>
    """
    _ATTRS_STYLE: EnumAdj_AttrAnnotsOrExisted = EnumAdj_AttrAnnotsOrExisted.ANNOTS_ONLY
    _ANNOTS_DEPTH: EnumAdj_AnnotsDepthAllOrLast = EnumAdj_AnnotsDepthAllOrLast.ALL_NESTED


# ---------------------------------------------------------------------------------------------------------------------
@final
class AttrAux_AnnotsLast(Base_AttrAux):
    """
    GOAL
    ----
    separate last/all nesting parents annotations
    """
    _ATTRS_STYLE: EnumAdj_AttrAnnotsOrExisted = EnumAdj_AttrAnnotsOrExisted.ANNOTS_ONLY
    _ANNOTS_DEPTH: EnumAdj_AnnotsDepthAllOrLast = EnumAdj_AnnotsDepthAllOrLast.LAST_CHILD


# =====================================================================================================================



import pytest

from base_aux.base_lambdas.m1_lambda import *
from base_aux.base_values.m4_primitives import *

from base_aux.valid.m3_valid_chains import *
from base_aux.valid.m2_valid_derivatives import *


# =====================================================================================================================
class Test__ValidChains:
    # -----------------------------------------------------------------------------------------------------------------
    @pytest.mark.parametrize(
        argnames="chains, _EXPECTED",
        argvalues=[
            ([True, ], True),
            ([False, ], False),
            ([None, ], False),

            ([0, ], False),
            ([1, ], True),      # CAREFUL assert 1 == True, assert 2 == False, assert 0 == False
            ([2, ], False),

            ([[], ], False),
            ([[None, ], ], False),

            ([Valid(True), ], True),
            ([Valid(False), ], False),
            ([Valid(False, skip_link=True), ], True),
            ([Valid(False, chain__cum=False), ], True),
        ]
    )
    def test__types_single(self, chains, _EXPECTED):
        func_link = ValidChains(chains).run
        Lambda(func_link).check_expected__assert(_EXPECTED)

    # -----------------------------------------------------------------------------------------------------------------
    @pytest.mark.parametrize(
        argnames="chains, _EXPECTED",
        argvalues=[
            ([ValidNoCum(False), ], True),
        ]
    )
    def test__nocum(self, chains, _EXPECTED):
        func_link = ValidChains(chains).run
        Lambda(func_link).check_expected__assert(_EXPECTED)

    # -----------------------------------------------------------------------------------------------------------------
    @pytest.mark.parametrize(
        argnames="chains, _EXPECTED",
        argvalues=[
            ([True, True, True], True),
            ([True, False, True], False),

            ([True, LAMBDA_TRUE, True], True),
            ([True, LAMBDA_TRUE, ClsCallTrue()], True),

            ([Valid(True), Valid(True)], True),
            ([Valid(True), Valid(False)], False),
            ([Valid(True), Valid(False, skip_link=True)], True),
            ([Valid(True), Valid(False, chain__cum=False)], True),

            ([True, ValidChains([True, True])], True),
            ([True, ValidChains([False, ], skip_link=True)], True),
            ([True, ValidChains([False, ], chain__cum=False)], True),
        ]
    )
    def test__chains(self, chains, _EXPECTED):
        func_link = ValidChains(chains).run
        Lambda(func_link).check_expected__assert(_EXPECTED)

    # -----------------------------------------------------------------------------------------------------------------
    @pytest.mark.parametrize(
        argnames="chains, _EXPECTED",
        argvalues=[
            ([ValidSleep(), True], True),
            ([ValidSleep(0.1), True], True),
            ([ValidSleep(0.1), False], False),

            ([ValidSleep(2, True), ], True),
        ]
    )
    def test__util1_sleep(self, chains, _EXPECTED):
        func_link = ValidChains(chains).run
        Lambda(func_link).check_expected__assert(_EXPECTED)

    @pytest.mark.parametrize(
        argnames="chains, _EXPECTED",
        argvalues=[
            ([False, ValidBreak(True), True], False),

            ([True, ValidBreak(True), False], True),
            ([True, ValidBreak(False), False], False),
            ([True, ValidBreak(False), True], True),

            ([ValidBreak(True), False], True),
            ([ValidBreak(False), False], False),
            ([ValidBreak(False), True], True),
        ]
    )
    def test__util2_break(self, chains, _EXPECTED):
        func_link = ValidChains(chains).run
        Lambda(func_link).check_expected__assert(_EXPECTED)

    # -----------------------------------------------------------------------------------------------------------------
    @pytest.mark.parametrize(
        argnames="chains",
        argvalues=[
            [True, True, True],
            [True, False, True],

            [True, LAMBDA_TRUE, True],
            [True, LAMBDA_TRUE, ClsCallTrue()],

            [Valid(True), Valid(True)],
            [Valid(True), Valid(False)],
            [Valid(True), Valid(False, skip_link=True)],
            [Valid(True), Valid(False, chain__cum=False)],

            [True, ValidChains([True, True])],
            [True, ValidChains([False, ], skip_link=True)],
            [True, ValidChains([False, ], chain__cum=False)],
        ]
    )
    def test__str(self, chains):
        assert str(ValidChains(chains)) is not None


# =====================================================================================================================



import time
import datetime


# =====================================================================================================================
class TimeoutWorker:   # starichenko
    # better use _LogProgressBar!
    def __init__(self, timeout=0):
        """
        :param timeout: if use 0 or None - used just like a stopwatch
        """
        self.time_start = time.time()
        self.timeout = timeout
        self._drop_flag = False

    def check_finish(self):
        return self._drop_flag or self.get_time_passed() > self.timeout

    def restart(self):
        self.time_start = time.time()
        self._drop_flag = False
        return True

    def wait_timeout(self, step=1, restart=False):
        """actually waiting whole timeout"""
        if restart:
            self.restart()
        while not self.check_finish():
            time.sleep(step)
        return True

    def drop(self):
        """stop timeout"""
        self._drop_flag = True
        return True

    # def get_time_passed(self, cutting_level=None):
    #     result = time.time() - self.time_start
    #     result = number_cutting(source=result, cutting_level=cutting_level)
    #     return result


def time_convert_to_human(seconds):
    # toDO: use standart python time formatting!
    try:
        if seconds <= 0:
            return ""
    except TypeError:
        raise Exception("Cannot convert '{}' to human time".format(seconds))

    human_time_string = ""
    days = seconds // 86400
    hours = seconds // 3600 % 24
    minutes = seconds // 60 % 60
    seconds = seconds % 60
    if days:
        human_time_string = human_time_string + str(days) + 'д.'
    if hours:
        human_time_string = human_time_string + str(hours) + 'ч.'
    if minutes:
        human_time_string = human_time_string + str(minutes) + 'м.'
    if seconds:
        human_time_string = human_time_string + str(seconds) + 'с.'

    return human_time_string


# def time_sleep_with_log(seconds=1, msg=None):   # starichenko
#     """ logging steps of pause sleep.
#     usefull when you want to be sure that process is going!
#     """
#     # debug_stack_get_list()
#     MSG_DEFAULT = f"WAIT seconds/step=[{seconds}]"
#     msg = msg or MSG_DEFAULT
#
#     seconds = float(seconds)
#     progress_obj = LogProgressBarTimeout(max_value=seconds, title=msg)
#
#     while not progress_obj.check_finish():
#         progress_obj.update()
#         time.sleep(1)
#
#     progress_obj.update_status(True)
#     return True


def time_sleep_remaining_time(seconds=1, start=None):   # starichenko
    """sleep remaining time from start point!

    :param seconds: expected time for sleep from start point
    :param start: start point time in seconds
        if None - sleep all seconds time

    :return: deviation seconds
        if deviation less then zero - return it without sleeping with minus!
    """
    time_now = time.time()

    if start is None:
        time_remain = seconds
    else:
        time_past = time_now - start
        time_remain = seconds - time_past

        if time_remain > 0:
            time.sleep(time_remain)

    return time_remain


def time_get_in_full_seconds_from_structtime(source=None):   # starichenko
    """
    created specially to get time in seconds to compare volga localtime with system.
    :param source:
    :return:
    """
    source = source or time.time()
    seconds = source.tm_sec + source.tm_min*60 + source.tm_hour*60*60
    return seconds


# =====================================================================================================================



from typing import *
import pytest

from base_aux.base_lambdas.m1_lambda import *
from base_aux.aux_text.m7_text_formatted import PatFormat, TextFormatted
from base_aux.versions.m2_version import Version


# =====================================================================================================================
class Test_Formatted:
    def test__pat_groups(self):
        assert PatFormat.SPLIT_STATIC__IN_PAT == r"(?:\{(?:[_a-zA-Z]\w*)?(?:[^{}]*)\})"

    def test__simple(self):
        victim = TextFormatted("{}", 1)
        assert victim.VALUES._0 == 1

        print("{}".format(1))
        print(str(victim))
        assert str(victim) == "1"

    def test__kwargs(self):
        # kwargs preferred ---------------------------------
        victim = TextFormatted("hello {name}={value}", *(1, 2))
        # assert victim.VALUES._1 == 1
        assert victim.VALUES.name == 1
        print(str(victim))
        assert str(victim) == "hello 1=2"

        victim.VALUES.name = "name"
        assert victim.VALUES.name == "name"
        print(str(victim))
        assert str(victim) == "hello name=2"

        # kwargs preferred ---------------------------------
        victim = TextFormatted("hello {name}={value}", "arg1", name="name", value=1)
        # assert victim.VALUES._1 == 1
        assert victim.VALUES.name == "name"
        print(str(victim))
        assert str(victim) == "hello name=1"

        victim.VALUES.name = "name2"
        assert victim.VALUES.name == "name2"
        print(str(victim))
        assert str(victim) == "hello name2=1"

        # args ---------------------------------
        victim = TextFormatted("hello {name}={value}", "arg1", value=1)
        # assert victim.VALUES._0 == "arg1"
        # assert victim.VALUES._1 == 1
        assert victim.VALUES.name == "arg1"
        print(str(victim))
        assert str(victim) == "hello arg1=1"

    def test__other(self):
        # OK --------
        victim = TextFormatted("hello {name}={value}", "arg1", value=1)
        # assert victim.VALUES._1 == 1
        assert victim.VALUES.name == "arg1"
        print(str(victim))
        assert str(victim) == "hello arg1=1"

        victim.other("hello name_other=222")
        print(str(victim))
        assert victim.VALUES.name == "name_other"
        assert victim.VALUES.value == 222

        victim.other("hello =222")
        print(str(victim))
        assert victim.VALUES.name == ""
        assert victim.VALUES.value == 222

        victim.other("hello  name_other=222")
        print(str(victim))
        assert victim.VALUES.name == " name_other"
        assert victim.VALUES.value == 222

        # EXC --------
        for wrong_value in ["he name_other=222", "hello name_222", "hello name_other222", ]:
            try:
                victim(wrong_value)
            except:
                pass
            else:
                print(111111, str(victim))
                print(222222, repr(victim))
                assert False

    # -----------------------------------------------------------------------------------------------------------------
    @pytest.mark.parametrize(
        argnames="pat_format,args,kwargs,_EXPECTED",
        argvalues=[
            ("hello {name}={value}", (), {}, "hello ="),
            ("hello {name}={value}", (), dict(name=1), "hello 1="),
            ("hello {name}={value}", (), dict(name=1, value=2), "hello 1=2"),
            ("hello {name}={value}", (11, 22), dict(name=1, value=2), "hello 1=2"),
            ("hello {name}={value}", (11, 22), dict(), "hello 11=22"),
            ("hello {name}={value}", (11, 22), dict(value=Version("1.2.3")), "hello 11=1.2.3"),

            ("hello\n\n{name}\n{value}", (11, 22), dict(), "hello\n\n11\n22"),

            ("hello {name}={value:3}", (), dict(name=1, value=2), "hello 1=  2"),
        ]
    )
    def test__str(self, pat_format, args, kwargs, _EXPECTED):
        func_link = lambda: str(TextFormatted(pat_format, *args, **kwargs))
        Lambda(func_link).check_expected__assert(_EXPECTED)

    # -----------------------------------------------------------------------------------------------------------------
    @pytest.mark.parametrize(
        argnames="pat_format,args,kwargs,_EXPECTED",
        argvalues=[
            ("hello{name}", (), dict(name=1), [int, 1]),
            ("hello{name}", (), dict(name="1"), [str, "1"]),
            ("hello{name}", (), dict(name=Version("1.2.3")), [Version, Version("1.2.3")]),
        ]
    )
    def test__type_apply(self, pat_format, args, kwargs, _EXPECTED):
        victim = TextFormatted(pat_format, *args, **kwargs)
        func_link = lambda: victim.VALUES.__annotations__["name"]
        Lambda(func_link).check_expected__assert(_EXPECTED[0])

        func_link = lambda: getattr(victim.VALUES, "name")
        Lambda(func_link).check_expected__assert(_EXPECTED[1])

        Lambda(lambda: victim.VALUES.name.__class__).check_expected__assert(_EXPECTED[0])

    # -----------------------------------------------------------------------------------------------------------------
    @pytest.mark.parametrize(
        argnames="pat_format,args,kwargs,new,_EXPECTED",
        argvalues=[
            ("hello{name}", (), dict(name=1), 1, [True, int, 1]),
            ("hello{name}", (), dict(name=1), "1", [True, int, 1]),
            ("hello{name}", (), dict(name=1), "text", [Exception, int, 1]),
            ("hello{name}", (), dict(name="1"), "text", [True, str, "text"]),
            ("hello{name}", (), dict(name="1"), "1", [True, str, "1"]),
            ("hello{name}", (), dict(name="1"), 1, [True, str, "1"]),
            ("hello{name}", (), dict(name=Version("1.2.3")), "1.2.3", [True, Version, Version("1.2.3")]),
        ]
    )
    def test__value_set(self, pat_format, args, new, kwargs, _EXPECTED):
        victim = TextFormatted(pat_format, *args, raise_types=True, **kwargs)

        # INCORRECT ------------------------------------
        # victim.VALUES.name = str(kwargs["name"])
        # func_link = lambda: victim.VALUES.name
        # Lambda(func_link).check_assert(_EXPECTED[1])
        #
        # Lambda(lambda: victim.VALUES.name.__class__).check_assert(_EXPECTED[0])

        # CORRECT ------------------------------------
        try:
            victim["name"] = new
            Lambda(True).check_expected__assert(_EXPECTED[0])
        except:
            Lambda(Exception).check_expected__assert(_EXPECTED[0])
            victim = TextFormatted(pat_format, *args, raise_types=False, **kwargs)
            victim["name"] = new    # NoRaise here!
            return
            pass

        Lambda(lambda: victim.VALUES.name.__class__).check_expected__assert(_EXPECTED[1])

        func_link = lambda: victim.VALUES.name
        Lambda(func_link).check_expected__assert(_EXPECTED[2])


# =====================================================================================================================



from typing import *

from base_aux.aux_attr.m1_annot_attr1_aux import *


# =====================================================================================================================
class NestInit_MutableClsValues:
    """
    GOAL
    ----
    reinit mutable values

    NOTE
    ----
    use simple copy past method!
    """
    def __init__(self, *args, **kwargs) -> None:
        AttrAux_Existed(self).reinit__mutable_cls_values()  # keep on first step!!! reinit only selfValues! not classvalues!
        super().__init__(*args, **kwargs)


# =====================================================================================================================
def test__reinit_mutables():
    # check common way ---------
    CONST = {}
    CONST2 = {1:1}

    assert {} == {}
    assert {} is not {}
    assert CONST is CONST

    assert {1:1} == {1:1}
    assert {1:1} is not {1:1}
    assert CONST2 is CONST2

    # NEW WAY -------------------
    class Victim:
        LIST = []
        SET = {}
        DICT = {}

    class Victim2(Victim, NestInit_MutableClsValues):
        pass

    victim = Victim()
    assert victim.LIST == Victim.LIST
    assert victim.LIST is Victim.LIST

    victim2 = Victim2()
    assert victim2.LIST == Victim2.LIST
    assert victim2.LIST is not Victim2.LIST


# =====================================================================================================================



from typing import *

import pandas as pd
import numpy as np


# =====================================================================================================================
TYPING__NP_TS__LINE = tuple[int,   float, float, float, float,   int, int, int]

TYPING__NP__FINAL = np.ndarray
TYPING__NP__DRAFT = np.ndarray | list[TYPING__NP_TS__LINE]

TYPING__NP_TS__FINAL = TYPING__NP__FINAL
TYPING__NP_TS__DRAFT = TYPING__NP__DRAFT

TYPING__PD_SERIES = pd.core.series.Series       # NOTE: dont use [pd.Series] it is not actually TYpe!=<class 'pandas.core.series.Series'>
TYPING__PD_DATAFRAME = pd.core.frame.DataFrame  # NOTE: dont use [pd.DataFrame] it is not actually TYpe!=<class 'pandas.core.frame.DataFrame'>


# =====================================================================================================================



import pytest

from base_aux.breeders.m1_breeder_str2_stack import *
from base_aux.breeders.m1_breeder_str1_series import *


# =====================================================================================================================
class Test__BreederStrStack:
    @classmethod
    def setup_class(cls):
        class BreederStrStack_Example(BreederStrStack):
            name0: int = 0
            name1: int = 1
            TAIL: BreederStrSeries = BreederStrSeries(2, 2, "%s")

        cls.victim = BreederStrStack_Example()
        pass

    # @classmethod
    # def teardown_class(cls):
    #     pass
    #
    # def setup_method(self, method):
    #     pass
    #
    #
    # def teardown_method(self, method):
    #     pass

    # -----------------------------------------------------------------------------------------------------------------
    @pytest.mark.parametrize(
        argnames="index_start",
        argvalues=[
            None,

            -2,
            -1,
            0,
            1,
            2,
        ]
    )
    def test__index_start(self, index_start):
        class Victim(BreederStrStack):
            _INDEX_START = index_start
            name0: int = None
            name1: int | None = None
            series23: BreederStrSeries = BreederStrSeries(None, 2, "series23_%s")
            series45: BreederStrSeries = BreederStrSeries(None, 2, "series45_%s")
            name6: None = None

        victim = Victim()

        index_start = index_start or 0
        assert victim[index_start +0] == "name0"
        assert victim[index_start +1] == "name1"
        assert victim[index_start +2] == "series23_1"
        assert victim[index_start +3] == "series23_2"
        assert victim[index_start +4] == "series45_1"
        assert victim[index_start +5] == "series45_2"
        assert victim[index_start +6] == "name6"

    def test__index_start__exc_overlayd(self):
        class Victim(BreederStrStack):
            name0: int = None
            name1: int | None = 0
            series23: BreederStrSeries = BreederStrSeries(None, 2, "series23_%s")
            series45: BreederStrSeries = BreederStrSeries(None, 2, "series45_%s")
            name6: None = None

        try:
            victim = Victim()
        except:
            pass
        else:
            assert False

    # -----------------------------------------------------------------------------------------------------------------
    def test__index__auto(self):
        class Victim(BreederStrStack):
            name0: int = 0
            name1: int | None = None
            series23: BreederStrSeries = BreederStrSeries(None, 2, "series23_%s")
            series45: BreederStrSeries = BreederStrSeries(None, 2, "series45_%s")
            name6: None = None

        victim = Victim()
        assert victim[0] == "name0"
        assert victim[1] == "name1"
        assert victim[2] == "series23_1"
        assert victim[3] == "series23_2"
        assert victim[4] == "series45_1"
        assert victim[5] == "series45_2"
        assert victim[6] == "name6"

        try:
            victim[7]
        except:
            pass
        else:
            assert False

    def test__index__break_order(self):
        class Victim(BreederStrStack):
            name0: int = 0
            name2: int = 2
            name1: int = 1

        victim = Victim()
        assert victim[0] == "name0"
        assert victim[1] == "name1"
        assert victim[2] == "name2"

        assert victim._DATA == {
            0: "name0",
            1: "name1",
            2: "name2",
        }

    # -----------------------------------------------------------------------------------------------------------------
    def test__exc_1__overlayd_1(self):
        class Victim(BreederStrStack):
            name0: int = 0
            name1: int = 1
            name2: int = 1
            TAIL: BreederStrSeries = BreederStrSeries(2, 2, "%s")

        try:
            victim = Victim()
        except:
            pass
        else:
            assert False

    def test__exc_1__overlayd_2(self):
        class Victim(BreederStrStack):
            name0: int = 0
            name1: int = 1
            TAIL: BreederStrSeries = BreederStrSeries(1, 2, "%s")

        try:
            victim = Victim()
        except:
            pass
        else:
            assert False

    def test__exc_1__skipped_1(self):
        class Victim(BreederStrStack):
            name0: int = 0
            name1: int = 1
            TAIL: BreederStrSeries = BreederStrSeries(2, 2, "%s")
            TAIL2: BreederStrSeries = BreederStrSeries(20, 2, "%s")

        try:
            victim = Victim()
        except:
            pass
        else:
            assert False

    def test__exc_1__skipped_2(self):
        class Victim(BreederStrStack):
            name0: int = 0
            # name1: int = 1
            TAIL: BreederStrSeries = BreederStrSeries(2, 2, "%s")

        try:
            victim = Victim()
        except:
            pass
        else:
            assert False

    # -----------------------------------------------------------------------------------------------------------------
    def test_1__main_1__getitem(self):
        assert self.victim[0] == "name0"
        assert self.victim[1] == "name1"
        assert self.victim[2] == "1"
        assert self.victim[3] == "2"
        try:
            assert self.victim[4]
        except:
            pass
        else:
            assert False

        assert self.victim["name0"] == 0
        assert self.victim["name1"] == 1
        assert self.victim["1"] == 2
        assert self.victim["2"] == 3
        try:
            assert self.victim["3"]
        except:
            pass
        else:
            assert False

    def test_1__main_2__contain(self):
        assert 0 in self.victim
        assert 1 in self.victim
        assert 2 in self.victim
        assert 3 in self.victim
        assert 4 not in self.victim
        assert 5 not in self.victim

        assert "name0" in self.victim
        assert "name1" in self.victim
        assert "attr11111" not in self.victim

        assert "0" not in self.victim
        assert "1" in self.victim
        assert "2" in self.victim
        assert "3" not in self.victim
        assert "4" not in self.victim
        assert "5" not in self.victim

    def test_2__sub_2__getitem(self):
        try:
            assert self.victim.TAIL[0]
        except:
            pass
        else:
            assert False

        assert self.victim.TAIL[1] == "1"
        assert self.victim.TAIL[2] == "2"
        try:
            assert self.victim.TAIL[3]
        except:
            pass
        else:
            assert False

        # 2--------
        try:
            assert self.victim.TAIL["0"]
        except:
            pass
        else:
            assert False
        assert self.victim.TAIL["1"] == 1
        assert self.victim.TAIL["2"] == 2
        try:
            assert self.victim.TAIL["3"]
        except:
            pass
        else:
            assert False

    def test_2__sub_2__contain(self):
        assert 0 not in self.victim.TAIL
        assert 1 not in self.victim.TAIL
        assert 2 in self.victim.TAIL
        assert 3 in self.victim.TAIL
        assert 4 not in self.victim.TAIL
        assert 5 not in self.victim.TAIL

        assert "0" not in self.victim.TAIL
        assert "1" in self.victim.TAIL
        assert "2" in self.victim.TAIL
        assert "3" not in self.victim.TAIL
        assert "4" not in self.victim.TAIL
        assert "5" not in self.victim.TAIL

    # -----------------------------------------------------------------------------------------------------------------
    def test__count(self):
        class Victim(BreederStrStack):
            name0: int = 0
            name1: int = 1
            TAIL: BreederStrSeries = BreederStrSeries(2, 2, "%s")

        assert Victim().count() == 4

        class Victim(BreederStrStack):
            name0: int = 0
            name1: int = 1
            TAIL: BreederStrSeries = BreederStrSeries(2, 2, "%s")
            TAIL2: BreederStrSeries = BreederStrSeries(4, 2, "%s")

        assert Victim().count() == 6


# =====================================================================================================================



import pytest

from base_aux.base_lambdas.m1_lambda import *

from base_aux.base_values.m3_exceptions import *
from base_aux.aux_attr.m3_ga1_prefix_1_inst import NestGa_Prefix_RaiseIf


# =====================================================================================================================
class Victim(NestGa_Prefix_RaiseIf):
    TRUE = True
    FALSE = False
    NONE = None

    def meth(self, value: Any = None):
        return value


# =====================================================================================================================
def test__register():
    # --------------------------------
    assert Victim().TRUE is True
    assert Victim().true is True

    # --------------------------------
    try:
        Victim().raise_if__true()
    except Exc__GetattrPrefix_RaiseIf:
        pass
    else:
        assert False
    try:
        Victim().raise_if__TRUE()
    except Exc__GetattrPrefix_RaiseIf:
        pass
    else:
        assert False
    try:
        Victim().RAISE_IF__TRUE()
    except Exc__GetattrPrefix_RaiseIf:
        pass
    else:
        assert False

    # --------------------------------
    assert Victim().raise_if_not__TRUE() is None
    assert Victim().raise_if_not__true() is None
    assert Victim().RAISE_IF_NOT__TRUE() is None


# =====================================================================================================================
def test__attr__not_exists():
    # TRUE ---------------------
    try:
        Victim().raise_if__HELLO()
    except Exception:
        pass
    else:
        assert False


# ---------------------------------------------------------------------------------------------------------------------
def test__attr__static():
    # TRUE ---------------------
    assert Victim().TRUE is True

    try:
        Victim().raise_if__TRUE()
    except Exc__GetattrPrefix_RaiseIf:
        pass
    else:
        assert False
    assert Victim().raise_if_not__TRUE() is None
    assert Victim().raise_if_not__true() is None
    assert Victim().RAISE_IF_NOT__TRUE() is None

    # FALSE ---------------------
    assert Victim().FALSE is False

    try:
        Victim().raise_if_not__FALSE()
    except Exc__GetattrPrefix_RaiseIf:
        pass
    else:
        assert False
    assert Victim().raise_if__FALSE() is None

    # NONE ---------------------
    assert Victim().NONE is None

    try:
        Victim().raise_if_not__NONE()
    except Exc__GetattrPrefix_RaiseIf:
        pass
    else:
        assert False
    assert Victim().raise_if__NONE() is None


# =====================================================================================================================
def test__meth_not_passed():
    # NOT_PASSED ---------------------
    assert Victim().meth() is None
    assert Victim().METH() is None

    try:
        Victim().METH2()
    except AttributeError:
        pass
    else:
        assert False


@pytest.mark.parametrize(
    argnames="args, _EXPECTED",
    argvalues=[
        (None, None),
        (True, Exc__GetattrPrefix_RaiseIf),
        (False, None),
    ]
)
def test___meth__raise_if(args, _EXPECTED):
    func_link = Victim().raise_if__METH
    Lambda(func_link, args).check_expected__assert(_EXPECTED)


@pytest.mark.parametrize(
    argnames="args, _EXPECTED",
    argvalues=[
        (None, Exc__GetattrPrefix_RaiseIf),
        (True, None),
        (False, Exc__GetattrPrefix_RaiseIf),
    ]
)
def test___meth__raise_if_not(args, _EXPECTED):
    func_link = Victim().raise_if_not__METH
    Lambda(func_link, args).check_expected__assert(_EXPECTED)


# =====================================================================================================================
# @pytest.mark.skip
def test__comment():
    COMMENT_APPLYED = "COMMENT_APPLYED"
    try:
        Victim().raise_if__METH(True, COMMENT=COMMENT_APPLYED)
    except Exc__GetattrPrefix_RaiseIf as exc:
        print(exc)
        # ObjectInfo(exc).print()
        assert COMMENT_APPLYED in str(exc)
    else:
        assert False


# =====================================================================================================================



from base_aux.aux_text.m1_text_aux import TextAux
from base_aux.path1_dir.m2_dir import *


try:
    import git  # GITPYTHON # need try statement! if not installed git.exe raise Exc even if module was setup!!!
except:
    print(f"[git.ERROR] is not setup in OS")


# =====================================================================================================================
class Git(DirAux):
    """
    GOAL
    ----
    get last commit short info instead of hard-version

    NOTE
    ----
    noraise
    """
    DIRPATH: TYPING.PATH_FINAL
    REPO: git.Repo = None                   # real object/only existed
    ROOT: TYPING.PATH_FINAL | None = None

    # -----------------------------------------------------------------------------------------------------------------
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._root_find()

    def _root_find(self) -> TYPING.PATH_FINAL | None:
        """
        GOAL
        ----
        detect/find exact root from passed Path (go parent recursively)
        """
        if not self.check_installed():
            return

        while True:
            try:
                self.REPO = git.Repo(self.DIRPATH)
                self.ROOT = self.DIRPATH
                print(f"[git.INFO] root detected {self.ROOT=}")
                return self.ROOT
            except git.InvalidGitRepositoryError:
                print(f"[git.WARN] root wrong {self.DIRPATH=}")
                parent = self.DIRPATH.parent
                if parent == self.DIRPATH:
                    return
                else:
                    self.DIRPATH = parent
                    continue
            except Exception as exc:
                print(f"[git.WARN] unexpected {exc!r}")

    # -----------------------------------------------------------------------------------------------------------------
    def check_installed(self) -> bool:
        """
        GOAL
        ----
        show that git is installed!

        NOTE
        ----
        need separate NOT check_ready(repo NOT created/wrong Path) and git not installed
        """
        try:
            import git
            return True
        except Exception as exc:
            print(f"[git.WARN] setup git! {exc!r}")
            return False

    def check_detected(self) -> bool:
        """
        GOAL
        ----
        check if all ready to work
        - git setup
        - root found
        - repo obj created
        """
        if self.REPO:
            return True
        else:
            return False

    # -----------------------------------------------------------------------------------------------------------------
    def check__status(self) -> bool:
        """
        GOAL
        ----
        check REPO exists + no untracked files + state is not DIRTY (no uncommited changes in indexed files)
        """
        return self.DIRTY is False and not self.UNTRACKED_FILES

    @property
    def DIRTY(self) -> bool | None:
        """
        GOAL
        ----
        check have uncommited changes!
        ONLY CHANGES IN INDEXED FILES!!!
        """
        if self.check_detected():
            return self.REPO.is_dirty()

    @property
    def UNTRACKED_FILES(self) -> list[str] | None:
        """
        GOAL
        ----
        return list NOT INDEXED files!
        """
        if self.check_detected():
            return self.REPO.untracked_files

    # -----------------------------------------------------------------------------------------------------------------
    @property
    def LIST_BRANCHES(self) -> list[git.Head]:
        """
        GOAL
        ----
        get all branch names
        """
        return [*self.REPO.branches]

    def list_commits(self, branch_name: str, limit: int = 10) -> list[git.Head]:
        """
        GOAL
        ----
        get all branch names
        """
        return [branch for branch in self.REPO.branches]

    # -----------------------------------------------------------------------------------------------------------------
    @property
    def COMMITTER(self) -> str | None:
        """
        EXAMPLE
        -------
        ndrei Starichenko
        """
        if self.check_detected():
            return self.REPO.head.object.committer

    @property
    def BRANCH(self) -> str | None:
        """
        EXAMPLE
        -------
        main
        """
        if self.check_detected():
            try:
                result = self.REPO.active_branch.name
            except Exception as exc:
                msg = f"[GIT] DETACHED HEAD - you work not on last commit on brange! {exc!r}"
                print(msg)
                result = "*DETACHED_HEAD*"
            return result

    @property
    def SUMMARY(self) -> str | None:
        """
        actual commit text

        EXAMPLE
        -------
        [Text] add shortcut_nosub
        """
        if self.check_detected():
            return self.REPO.commit().summary

    @property
    def HEXSHA(self) -> str | None:
        """
        NOTE
        ----
        see other more useful work with 8 chars! that's enough!

        EXAMPLE
        -------
        9fddeb5a9bed20895d56dd9871a69fd9dee5fbf7
        """
        if self.check_detected():
            return self.REPO.head.object.hexsha

    @property
    def HEXSHA8(self) -> str | None:
        """
        GOAL
        ----
        just a short variant for main HEXSHA (cut by last 8 chars)

        EXAMPLE
        -------
        9fddeb5a
        """
        if self.check_detected():
            return self.HEXSHA[:8]

    @property
    def DATETIME(self) -> datetime.datetime | None:
        """
        EXAMPLE
        -------
        2024-12-05 11:30:17+03:00
        """
        if self.check_detected():
            return self.REPO.head.object.committed_datetime

    # -----------------------------------------------------------------------------------------------------------------
    def git_mark__get(self) -> str:
        """
        EXAMPLE
        -------
        git_mark='[git_mark//main/zero/Andrei Starichenko/ce5c3148/2024-12-04 18:39:10]'
        """
        if self.check_detected():
            dirty = "!DIRTY!" if self.DIRTY else ""
            untrachked = "!UNTR!" if self.UNTRACKED_FILES else ""
            branch = TextAux(self.BRANCH).shortcut(15)
            summary = TextAux(self.SUMMARY).shortcut(15)
            dt = TextAux(self.DATETIME).shortcut_nosub(19)

            result = f"{dirty}{untrachked}{branch}/{summary}/{self.COMMITTER}/{self.HEXSHA8}/{dt}"

        else:
            result = f"вероятно GIT не установлен"

        git_mark = f"[git_mark//{result}]"
        print(f"{git_mark=}")
        return git_mark

    # -----------------------------------------------------------------------------------------------------------------
    # -----------------------------------------------------------------------------------------------------------------
    # -----------------------------------------------------------------------------------------------------------------
    # -----------------------------------------------------------------------------------------------------------------
    # -----------------------------------------------------------------------------------------------------------------
    # -----------------------------------------------------------------------------------------------------------------
    def is_commit_latest(self) -> bool:
        """
        GOAL
        ----
        check if commit is latest
        """

    def pull(self) -> bool:
        """
        GOAL
        ----
        get all updates from server!
        """


# =====================================================================================================================
if __name__ == '__main__':
    victim = Git()
    print()
    print(victim.git_mark__get())
    # ObjectInfo(victim.REPO).print()


# =====================================================================================================================



from typing import *
import re


# =====================================================================================================================
class CmdArgsKwargsParser:
    """
    GOAL
    ----
    parse string line for Cmd with args Kwargs with syntax
        "prefix cmdName arg1 arg2 kwarg1=val1 kwarg2=val2"
    get exact values prefix/CMD/Args/Kwargs

    NOTE
    ----
    ALL RESULTS IN LOWERCASE! (EXCEPT SOURCE!)

    SPECIALLY CREATED FOR
    ---------------------
    buses.serialClient
    """
    # INITS ------------
    SOURCE: str

    # RESULTS ------------
    PREFIX: str = ""
    CMD: str = ""

    ARGS: list[str]
    KWARGS: dict[str, str]

    # -----------------------------------------------------------------------------------------------------------------
    def __init__(self, source: str, prefix_expected: str = None):
        self.SOURCE = str(source)

        self.ARGS = []
        self.KWARGS = {}

        self.parse(prefix_expected=prefix_expected)

    # -----------------------------------------------------------------------------------------------------------------
    def parse(self, prefix_expected: str = None) -> None:
        # SOURCE fix ---------
        source = self.SOURCE.lower()
        source = re.sub(r"\s*=+\s*", "=", source)

        # PREFIX -------------
        prefix_expected = prefix_expected or ""
        prefix_expected = prefix_expected.lower()

        if prefix_expected and source.startswith(prefix_expected):
            self.PREFIX = prefix_expected

            source = source.replace(self.PREFIX, "", 1)

        # --------------------
        splits = source.split()
        if not splits:
            return

        # ARGS/KWARGS ----------------
        for part in splits:
            if "=" not in part:
                if not self.CMD:
                    self.CMD = part
                else:
                    self.ARGS.append(part)
            else:
                part__key_value = part.split("=")
                self.KWARGS.update(dict([part__key_value, ]))


# =====================================================================================================================



from typing import *
import pytest

from base_aux.base_lambdas.m1_lambda import *
from base_aux.base_lambdas.m1_lambda2_derivatives import *
from base_aux.aux_dict.m3_dict_ga import *


# =====================================================================================================================
dict_example = {
    "lowercase": "lowercase",
    "nested": {"n1":1},
}


class Victim(DictIcKeys_Ga):
    lowercase: str


# =====================================================================================================================
def test__init():
    victim = Victim()
    assert victim == {}

    Lambda(lambda: victim[1]).check_expected__assert(Exception)
    Lambda(lambda: victim.lowercase).check_expected__assert(Exception)

    victim = Victim({})
    assert victim == {}
    Lambda(lambda: victim[1]).check_expected__assert(Exception)
    Lambda(lambda: victim.lowercase).check_expected__assert(Exception)

    victim = Victim(**{})
    assert victim == {}
    Lambda(lambda: victim[1]).check_expected__assert(Exception)
    Lambda(lambda: victim.lowercase).check_expected__assert(Exception)

    victim = Victim({1:1})
    assert victim == {1:1}
    Lambda(lambda: victim[1]).check_expected__assert(1)
    Lambda(lambda: victim.lowercase).check_expected__assert(Exception)

    victim = Victim(dict_example)
    assert victim == dict_example
    Lambda(lambda: victim[1]).check_expected__assert(Exception)
    Lambda(lambda: victim.lowercase).check_expected__assert("lowercase")

    victim = Victim(**dict_example)
    assert victim == dict_example
    Lambda(lambda: victim[1]).check_expected__assert(Exception)
    Lambda(lambda: victim.lowercase).check_expected__assert("lowercase")


# ---------------------------------------------------------------------------------------------------------------------
def test__not_exist():
    victim = Victim(dict_example)
    assert victim == dict_example

    Lambda(lambda: victim.NOT_EXIST).check_expected__assert(Exception)
    Lambda(lambda: victim["NOT_EXIST"]).check_expected__assert(Exception)
    victim["NOT_EXIST"] = 1
    assert victim.not_exist == 1
    assert victim["not_exist"] == 1


# ---------------------------------------------------------------------------------------------------------------------
def test__access_get():
    victim = Victim(dict_example)
    assert victim == dict_example
    assert victim.get("lowercase") == "lowercase"
    assert victim.get("LOWERCASE") == "lowercase"


def test__access_attrs():
    victim = Victim(dict_example)
    assert victim == dict_example

    assert victim.lowercase == "lowercase"
    assert victim.LOWERCASE == "lowercase"

    victim.lowercase = 111
    assert victim.lowercase == 111
    assert victim.LOWERCASE == 111

    victim.LOWERCASE = 222
    assert victim.lowercase == 222
    assert victim.LOWERCASE == 222


def test__access_items():
    victim = Victim(dict_example)
    assert victim == dict_example

    assert victim["lowercase"] == "lowercase"
    assert victim["LOWERCASE"] == "lowercase"

    victim["lowercase"] = 111
    assert victim["lowercase"] == 111
    assert victim["LOWERCASE"] == 111

    victim["LOWERCASE"] = 222
    assert victim["lowercase"] == 222
    assert victim["LOWERCASE"] == 222
    victim["lowercase"] = "lowercase"

    Lambda(lambda: victim[1]).check_expected__assert(Exception)
    # assert victim[1] == None
    victim[1] = 1
    assert victim[1] == 1


def test__nested():
    victim = Victim(dict_example)

    assert victim.nested == {"n1":1}
    assert victim.nested.n1 == 1

    victim.nested.n1 = 2
    assert victim.nested.n1 != 2    # NESTED LEVELS ONLY READ ONLY!!!
    assert victim.nested.n1 == 1


# ---------------------------------------------------------------------------------------------------------------------
def test__del():
    victim = Victim()

    victim.hello = None
    assert "hello" in victim
    del victim["hello"]
    assert "hello" not in victim

    victim.hello = None
    assert "hello" in victim
    del victim.hello
    assert "hello" not in victim


def test__pop():
    victim = Victim()

    victim.hello = None
    assert "hello" in victim
    victim.pop("hello")
    assert "hello" not in victim

    victim.hello = None
    assert "hello" in victim
    victim.pop("HELLO")
    assert "hello" not in victim


def test__clear():
    victim = Victim({1:1})
    assert victim == {1:1}
    victim.clear()
    assert victim == {}


# ---------------------------------------------------------------------------------------------------------------------
def test__contain():
    victim = Victim(dict_example)
    assert victim == dict_example

    assert 1 not in victim
    assert "HELLO" not in victim
    assert "lowercase" in victim
    assert "LOWERCASE" in victim


def test__list_len_bool():
    victim = Victim()
    assert list(victim) == []
    assert len(victim) == 0
    assert bool(victim) == False
    victim[1] = 1
    assert list(victim) == [1]
    assert len(victim) == 1
    assert bool(victim) == True


def test__iter():
    dict_1 = {1:1, "hello": 2}
    victim = Victim(dict_1)
    for key1, key2 in zip(victim, dict_1):
        assert key1 == key2


def test__update():
    victim = Victim()
    victim[1] = 1
    victim.update({1:11})
    assert victim[1] == 11


# =====================================================================================================================



from typing import *
import pytest

from base_aux.aux_attr.m4_kits import *
from base_aux.aux_attr.m5_attr_diff import *

from base_aux.base_lambdas.m1_lambda import *
from base_aux.aux_attr.m4_dump1_dumping1_dict import *
from base_aux.aux_attr.m0_static import *

from base_aux.base_values.m2_value_special import *


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="source1, source2, _EXPECTED",
    argvalues=[
        (
                AttrKit_Blank(),
                AttrKit_Blank(),
                [
                    {},
                    {},
                    {},
                ]
        ),
        (
                ExampleAttrs1_Existed(),
                ExampleAttrs1_Existed(),
                [
                    {},
                    {},
                    {},
                ]
        ),
        (
                ExampleAttrs1_Existed(),
                ExampleAttrs21_AnnotMiddle(),
                [
                    {
                        'AN2': (NoValue, 2), '_AN2': (NoValue, 22), 'meth2': (NoValue, 2), '_meth2': (NoValue, 22),
                    },
                    {
                        'AN2': (NoValue, 2), '_AN2': (NoValue, 22),
                    },
                    {
                        'AN2': (NoValue, 2), '_AN2': (NoValue, 22),
                    },
                ]
        ),
        (
                ExampleAttrs1_Existed(),
                ExampleAttrs321_AnnotLast(),
                [
                    {
                        'AN2': (NoValue, 2), '_AN2': (NoValue, 22), 'meth2': (NoValue, 2), '_meth2': (NoValue, 22),
                        'AN3': (NoValue, 3), '_AN3': (NoValue, 33), 'meth3': (NoValue, 3), '_meth3': (NoValue, 33),
                    },
                    {
                        'AN2': (NoValue, 2), '_AN2': (NoValue, 22),
                        'AN3': (NoValue, 3), '_AN3': (NoValue, 33),
                    },
                    {
                        'AN3': (NoValue, 3), '_AN3': (NoValue, 33),
                    },
                ]
        ),
        (
                ExampleAttrs21_AnnotMiddle(),
                ExampleAttrs321_AnnotLast(),
                [
                    {
                        'AN3': (NoValue, 3), '_AN3': (NoValue, 33), 'meth3': (NoValue, 3), '_meth3': (NoValue, 33),
                    },
                    {
                        'AN3': (NoValue, 3), '_AN3': (NoValue, 33),
                    },
                    {
                        'AN2': (2, NoValue), '_AN2': (22, NoValue),
                        'AN3': (NoValue, 3), '_AN3': (NoValue, 33),
                    },
                ]
        ),
    ]
)
def test__names(source1, source2, _EXPECTED):
    Lambda(AttrDiff_Existed(source1, source2)).check_expected__assert(_EXPECTED[0])
    Lambda(AttrDiff_AnnotsAll(source1, source2)).check_expected__assert(_EXPECTED[1])
    Lambda(AttrDiff_AnnotsLast(source1, source2)).check_expected__assert(_EXPECTED[2])


# =====================================================================================================================



from typing import *
from base_aux.base_values.m3_exceptions import *


# =====================================================================================================================
class BreederStrSeries(NamedTuple):
    """
    PATTERN FOR BREEDING ONE TYPE OF TEMPLATE STYLE
    used and created special for BreederStrStack
    """
    START_OUTER: int | None  # None used only for recreation in Stack!!! for auto
    COUNT: int
    TEMPLATE: str = "%s"
    START_INNER: int = 1    # just a starting number in values!

    def _raise_if_start_outer_none(self) -> NoReturn | None:
        if self.START_OUTER is None:
            raise Exc__StartOuterNONE_UsedInStackByRecreation()

    def __contains__(self, item: Union[int, str]) -> Union[bool, NoReturn]:
        """
        :param item: one of: 1=OUTER_INDEX (not self-INNER!!!) or 2=ORIGINAL value_NAME
            couse of we usually need to compare OUTER indexes or original VALUES!
        :return:
        """
        self._raise_if_start_outer_none()

        if isinstance(item, int):
            return self.START_OUTER <= item < (self.START_OUTER + self.COUNT)

        elif isinstance(item, str):
            return item in self.get_dict__outer().values()

    def __getitem__(self, item: Union[int, str]) -> Union[int, str, NoReturn]:
        """

        :param item: INTERNAL INDEX! NOT OUTER!!!
        :return: INTERNAL NAME
        """
        _DATA = self.get_dict__inner()

        if item in _DATA:
            return _DATA[item]

        if item in _DATA.values():
            for key, value in _DATA.items():
                if item == value:
                    return key

        msg = f"{item=}"
        raise Exc__NotExistsNotFoundNotCreated(msg)

    def get_dict__outer(self) -> dict[int, str] | NoReturn:
        self._raise_if_start_outer_none()

        result_outer_sub = {}
        for index in range(self.COUNT):
            pos = index + self.START_INNER
            value = self.TEMPLATE % pos

            index_outer = index + self.START_OUTER
            result_outer_sub.update({index_outer: value})
        return result_outer_sub

    def get_dict__inner(self) -> dict[int, str] | NoReturn:
        self._raise_if_start_outer_none()

        result_inner = {}
        for index in range(self.COUNT):
            pos = index + self.START_INNER
            value = self.TEMPLATE % pos

            result_inner.update({pos: value})
        return result_inner

    def get_listed_index__by_outer(self, index: int) -> int | NoReturn:
        """
        usually we have outer index and need to be able get from other list VALUE according to listed index fom this template!

        :param index:
        :return:
        """
        result = 0
        if index in self:
            for _key in self.get_dict__outer():
                if _key == index:
                    return result
                else:
                    result += 1
        else:
            raise Exc__NotExistsNotFoundNotCreated()

    def get_listed_index__by_value(self, value: str) -> int | NoReturn:
        result = 0
        if value in self:
            for _value in self.get_dict__outer().values():
                if _value == value:
                    return result
                else:
                    result += 1
        else:
            raise Exc__NotExistsNotFoundNotCreated()


# =====================================================================================================================



from typing import *

from dataclasses import dataclass, field


# =====================================================================================================================
class ItemKeyValue(NamedTuple):
    KEY: str
    VALUE: str


# =====================================================================================================================
@dataclass
class ObjectState:
    """
    GOAL
    ----
    keep final results for obj attributes (call if callable)
    """
    # TODO: add sort method!!!???
    SKIPPED_BUILDIN: list[str] = field(default_factory=list)
    SKIPPED_FULLNAMES: list[str] = field(default_factory=list)
    SKIPPED_PARTNAMES: list[str] = field(default_factory=list)

    PROPERTIES__ELEMENTARY_SINGLE: dict[str, Any] = field(default_factory=dict)
    PROPERTIES__ELEMENTARY_COLLECTION: dict[str, Any] = field(default_factory=dict)
    PROPERTIES__OBJECTS: dict[str, Any] = field(default_factory=dict)
    PROPERTIES__EXC: dict[str, Exception] = field(default_factory=dict)

    METHODS__ELEMENTARY_SINGLE: dict[str, Any] = field(default_factory=dict)
    METHODS__ELEMENTARY_COLLECTION: dict[str, Any] = field(default_factory=dict)
    METHODS__OBJECTS: dict[str, Any] = field(default_factory=dict)
    METHODS__EXC: dict[str, Exception] = field(default_factory=dict)

    def items(self) -> Iterable[tuple[str, Union[list[str],  dict[str, Any]]]]:
        for group_name, group_values in self.__getstate__().items():
            yield group_name, group_values


# =====================================================================================================================



from base_aux.loggers.m1_print import *
from base_aux.testplans.devices_base import *
from base_aux.breeders.m2_table_inst import *
from base_aux.testplans.TESTPLANS.DEVICES import atc, ptb
from base_aux.testplans.devices_kit import DeviceKit


# =====================================================================================================================
class DeviceLines__AtcPtbDummy(DeviceKit):
    ATC = TableLine(atc.DeviceDummy(3))
    DUT = TableLine(*[ptb.DeviceDummy(index) for index in range(2)])


# =====================================================================================================================
class DeviceLines__Psu800(DeviceKit):
    ATC = TableLine(atc.Device(3))
    DUT = TableLine(*[ptb.Device(index) for index in range(10)])

    def resolve_addresses(self) -> None:
        class Dev(SerialClient):
            BAUDRATE = 115200
            EOL__SEND = b"\n"

        result = Dev.addresses_dump__answers("*:get name", "*:get addr")
        print(result)
        for port, responses in result.items():
            name_i = responses["*:get name"]
            addr_i = responses["*:get addr"]
            Print(f"{port}{responses}")

            if name_i == "ATC":
                filter_link = lambda dev: dev.NAME == name_i
            elif name_i == "PTB":
                filter_link = lambda dev: dev.NAME == name_i and dev.INDEX+1 == addr_i
            else:
                continue

            match = list(filter(filter_link,  self.iter_lines_insts()))
            dev_found = match and list(match)[0]
            if dev_found:
                dev_found.ADDRESS = port

        for dev in self.iter_lines_insts():
            if not isinstance(dev.ADDRESS, str):
                dev.ADDRESS = Enum__AddressAutoAcceptVariant.NOT_FOUND

        pass


# =====================================================================================================================
if __name__ == '__main__':
    DeviceLines__Psu800().resolve_addresses()


# =====================================================================================================================



import pathlib
from base_aux.path2_file.m1_filepath import *
from base_aux.path2_file.m2_file import *
from base_aux.path2_file.m3_filetext import *


# =====================================================================================================================
text_load = "text_load"
FILEPATH = Resolve_FilePath(namefull="victim.txt").resolve()


# =====================================================================================================================
class Test__textFile:
    def setup_class(cls):
        pass

    @classmethod
    def teardown_class(cls):
        pass

    def setup_method(self, method):
        pass

    def teardown_method(self, method):
        FileAux(filepath=FILEPATH).delete_file()

    # -----------------------------------------------------------------------------------------------------------------
    def test__File(self):
        self.victim = FileAux(filepath=FILEPATH)

        assert self.victim.clear_file() is True
        assert self.victim.read__text() == ""
        assert self.victim.TEXT == ""
        assert self.victim.write__text(text_load) == 9
        assert self.victim.TEXT == text_load
        assert self.victim.delete_file() is True

        assert self.victim.append__lines(1, 2) == True
        assert self.victim.TEXT == "1\n2"
        assert self.victim.append__lines(3, 4) == True
        assert self.victim.TEXT == "1\n2\n3\n4"
        assert self.victim.delete_file() is True

    def test__textFile(self):
        self.victim = TextFile(filepath=FILEPATH, text=123)
        assert self.victim.write__text() == 3
        assert self.victim.TEXT == "123"
        assert self.victim.delete_file() is True


# =====================================================================================================================



from base_aux.alerts.m1_alerts1_smtp import *
from base_aux.alerts.m1_alerts2_telegram import *
from base_aux.base_lambdas.m1_lambda import *


# =====================================================================================================================
# TODO: separate by base class and use victimCls as attr!

@pytest.mark.skip
@pytest.mark.skipif(Lambda(AlertSmtp.CONN_AUTH).check_raised__bool(), reason="no file")
class Test__1:
    @pytest.mark.parametrize(argnames="cls", argvalues=[
        AlertSmtp,
        AlertTelegram,
    ])
    def test__send_single(self, cls):
        victim = cls()
        victim.send_msg("single")
        assert victim.wait()

    @pytest.mark.parametrize(argnames="victim", argvalues=[
        # AlertSmtp,
        # AlertTelegram,
    ])
    @pytest.mark.parametrize(argnames="subject, msg, _subtype", argvalues=[
        (None, "zero", None),
        ("", "plain123", "plain123"),
        ("plain", "plain", "plain"),
        ("html", "<p><font color='red'>html(red)</font></p>", "html")
    ])
    def test__send_single__parametrized(self, victim, subject, msg, _subtype):
        victim = victim()
        victim.send_msg(subject=subject, body=msg, _subtype=_subtype)
        assert victim.result_sent_all is not True
        victim.wait()
        assert victim.result_sent_all is True

    @pytest.mark.parametrize(argnames="victim", argvalues=[
        AlertSmtp,
        # AlertTelegram
    ])
    def test__send_multy__result_wait(self, victim):
        pass
        # assert victim("multy1").wait() is True
        # assert victim("multy2").wait() is True

    @pytest.mark.parametrize(argnames="victim", argvalues=[
        AlertSmtp,
        # AlertTelegram
    ])
    def test__send_multy__wait_join(self, victim):
        return

        thread1 = victim("thread1")
        thread2 = victim("thread2")

        thread1.wait()
        thread2.wait()

        assert thread1._result is True
        assert thread2._result is True


# =====================================================================================================================
# class Test_AlertHttp:
#     Victim = AlertHttp
#     def test__send_single(self, victim):
#         assert self.Victim({}).result_wait() is True


# =====================================================================================================================



from typing import *


# =====================================================================================================================
class NestInit_Post:
    """
    GOAL
    ----
    use in all variants - show explicitly
    """
    def __init__(self, *args, **kwargs) -> None | NoReturn:
        super().__init__(*args, **kwargs)

        self._init_post()    # call always last!!!

    def _init_post(self) -> None | NoReturn:     # call always last!!!
        """
        GOAL
        ----
        user initions and checks

        TYPICAL USAGE
        -------------
        make some conversations for source, like str for text
        or
        make initial tests/checks for source, like typecheck
        """
        return NotImplemented


# =====================================================================================================================



import datetime

from base_aux.base_lambdas.m1_lambda import *
from base_aux.valid.m2_valid_derivatives import *
from base_aux.base_enums.m2_enum1_adj import *


# =====================================================================================================================
TYPING__CHAINS = list[Union[Valid, 'ValidChains', Any]]      # all Any will be converted to Valid!


# =====================================================================================================================
class ValidChains(Valid):
    """
    GOAL
    ----

    CREATED SPECIALLY FOR
    ---------------------

    CONSTRAINTS
    -----------

    BEST USAGE
    ----------
    val_chains = ValidChains(
        chains=[
            True,
            1+1 == 2,
            Valid(2),
            Valid(3, chain_cum=False),
            ValidChains([Valid(21), Valid(22)], chain_cum=False),
        ]
    )

    result = val_chains.run()

    WHY NOT: 1?
    -----------

    WHY NOT: 2?
    -----------
    """
    _CHAINS: TYPING__CHAINS

    def __init__(self, chains: TYPING__CHAINS, **kwargs):
        super().__init__(value_link=None, **kwargs)
        self._CHAINS = chains

    def __len__(self) -> int:
        return len(self._CHAINS)

    def __iter__(self):
        return iter(self._CHAINS)

    def run(self) -> bool:
        self.clear()
        self.timestamp_last = datetime.datetime.now()

        # SKIP ---------------------
        self.skip_last = Lambda(self.SKIP_LINK).resolve__bool()

        if not self.skip_last:
            # WORK =======================
            self.STATE_ACTIVE = EnumAdj_ProcessStateActive.STARTED
            self.log_lines.append(f"(START) len={len(self)}/timestamp={self.timestamp_last}")

            # init self.validate_last if None -----------
            # if self.validate_last is None:
            #     self.validate_last = True

            # ITER -----------
            for index, step in enumerate(self):
                if not isinstance(step, (Valid, ValidChains)):
                    step = Valid(step)

                step_result = step.run()
                self.log_lines.append(str(step))

                if step.skip_last:
                    continue

                if step.CHAIN__CUM:
                    self.validate_last &= step_result
                    if step.CHAIN__FAIL_STOP and not step_result:
                        self.log_lines.append(f"(FAIL STOP) [result={bool(self)}]{index=}/len={len(self)}")
                        break

                if isinstance(step, ValidBreak) and step_result:
                    break

            # -----------------
            self.STATE_ACTIVE = EnumAdj_ProcessStateActive.FINISHED
            self.log_lines.append(f"(FINISH) [result={bool(self)}]/len={len(self)}")    # need after finish! to keep correct result
            # ============================

        return bool(self)


# =====================================================================================================================
if __name__ == "__main__":
    victim = ValidChains([Valid(True), Valid(False)])
    print(victim)
    print()

    victim.run()
    print(victim)


# =====================================================================================================================



from typing import *
import time
from collections import deque

from PyQt5.QtCore import QThread, QTimer

from base_aux.base_nest_dunders.m3_calls import *
from base_aux.aux_text.m7_text_formatted import *
from base_aux.base_nest_dunders.m1_init0_reinit2_lambdas_resolve import NestInit_AttrsLambdaResolve


# =====================================================================================================================
class Interface_Alert:
    """Interface for Alerts

    RULES:
    - if some method cant exists (like for telegram) - just return True!
    - Dont return None!
    - Dont use Try sentences inside - it will be applied in upper logic!
    - Decide inside if it was success or not, and return conclusion True/False only.
    """
    MSGS_UNSENT: deque[str | TextFormatted | Any]
    MSG_ACTIVE: Any

    def _connect_unsafe(self) -> Union[bool, NoReturn]:
        """establish connection to source
        """
        # self._conn = None
        return True

    def _disconnect_unsafe(self) -> Union[bool, NoReturn]:
        """establish connection to source
        """
        return True

    def _login_unsafe(self) -> Union[bool, NoReturn]:
        """authorise to source
        """
        return True

    def _send_unsafe(self) -> Union[bool, NoReturn]:
        """send msg
        """
        return True

    def _msg_compose(self) -> Union[str, 'MIMEMultipart']:
        """generate msg from existed data in attributes (passed before on init)
        """
        return str(self.MSG_ACTIVE)

    def _recipient_get(self) -> str:
        """RECIPIENT SelfSending, get from obvious class base_types!
        """
        pass


# =====================================================================================================================
class Base_Alert(NestInit_AttrsLambdaResolve, Interface_Alert, NestCall_MethodName, QThread):     # REM: DONT ADD SINGLETON!!! SNMP WILL NOT WORK!!! and calling logic will be not simple!
    """
    GOAL
    ----
    alert msg sender
    """
    # SETTINGS ------------------------------------
    CONN_ADDRESS: AttrKit_Blank = None
    CONN_AUTH: AttrKit_AuthNamePwd = None
    RECIPIENT: Any = None

    TIMER_TIMEOUT:  int = 60
    TIMEOUT_SEND: float = 1.2
    RECONNECT_LIMIT: int = 3
    RECONNECT_PAUSE: int = 60
    # TIMEOUT_RATELIMIT: int = 600    # when EXC 451, b'Ratelimit exceeded

    MSGS_UNSENT: deque[str | TextFormatted | Any]

    # AUX -----------------------------------------
    _CALL__METHOD_NAME = "send_msg"

    _conn: Any = None
    result_connect: bool | None = None
    result_login: bool | None = None
    result_sent_all: bool | None = None

    timer: QTimer

    @property
    def MSG_ACTIVE(self) -> str | TextFormatted | Any | None:
        if len(self.MSGS_UNSENT) > 0:
            return self.MSGS_UNSENT[0]

    # =================================================================================================================
    def __init__(self, conn_address: AttrKit_Blank = None, conn_auth: AttrKit_AuthNamePwd = None, recipient: Any = None):
        """
        GOAL
        ----
        Send msg on init
        """
        # params --------------
        if conn_address is not None:
            self.CONN_ADDRESS = conn_address
        if conn_auth is not None:
            self.CONN_AUTH = conn_auth

        if recipient is not None:
            self.RECIPIENT = recipient
        if self.RECIPIENT is None:
            self.RECIPIENT = self._recipient_get()

        self.MSGS_UNSENT = deque()

        super().__init__()
        self.init_timer()

        # START ---------------
        self.start()

    def init_timer(self) -> None:
        self.timer = QTimer()
        self.timer.setInterval(self.TIMER_TIMEOUT * 1000)
        self.timer.timeout.connect(self.start)
        self.timer.start()

    # =================================================================================================================
    def send_msg(self, body):
        self.MSGS_UNSENT.append(body)
        self.start()

    def start(self, **kwargs):
        if not self.isRunning():
            super().start()

    # =================================================================================================================
    def _connect(self) -> Optional[bool]:
        """create connection object
        """
        while True:
            counter = 0
            self.ready_to_send = None
            while counter <= self.RECONNECT_LIMIT:
                counter += 1

                if not self.result_connect:
                    print(f"[connect] TRY {self.__class__.__name__}")
                    try:
                        self.result_connect = self._connect_unsafe()
                        if self.result_connect:
                            print("[connect] SUCCESS")
                    except Exception as exc:
                        print(f"[connect] ERROR [{exc!r}]")

                if self.result_connect and not self.result_login:
                    try:
                        self.result_login = self._login_unsafe()
                        if self.result_login:
                            print("[login] SUCCESS")
                    except Exception as exc:
                        print(f"[LOGIN] ERROR [{exc!r}]")

                self.ready_to_send = self.result_connect and self.result_login
                if self.ready_to_send:
                    break
                else:
                    print("=" * 100)
                    print("=" * 100)
                    print("=" * 100)
                    time.sleep(self.RECONNECT_PAUSE)

            # -------------------------------------------
            if self.ready_to_send:
                break
            else:
                print(f"RECONNECT_PAUSE[{counter=}]")
                print("=" * 100)
                print()
                time.sleep(10)
                continue

        return self.ready_to_send

    def _disconnect(self):
        self.result_connect = None
        self.result_login = None

        try:
            self._disconnect_unsafe()
        except:
            pass

    # =================================================================================================================
    def run(self) -> None:
        print()
        print()
        print()
        print(f"RUN START")
        self.result_sent_all = None

        while self.MSG_ACTIVE is not None:
            if not self._connect():
                break

            MSG = self._msg_compose()
            print("[Try send", "-" * 80)
            print(MSG)
            print("Try send]", "-" * 80)

            try:
                sent_result = self._send_unsafe()
                if sent_result:
                    self.MSGS_UNSENT.popleft()
                    print("[send] SUCCESS")
            except Exception as exc:
                msg = f"[sent] ERROR [{exc!r}]"
                # [send] ERROR [SMTPDataError(451, b'Ratelimit exceeded for mailbox centroid@mail.ru. Try again later.')]
                print(msg)
                break

        self.result_sent_all = self.MSG_ACTIVE is None
        print(f"{self.result_sent_all=}")
        self._disconnect()
        print(f"RUN FINISH")


# =====================================================================================================================



# FIXME: USE Base_KwargsEqExpect instead!!!!???? or decide to do smth with it!


from typing import *

from base_aux.base_values.m2_value_special import NoValue
from base_aux.aux_eq.m3_eq_valid1_base import *


# =====================================================================================================================
class Meta__Enum(type):
    """
    GOAL
    ----
    1/ remake ITEMS to exact EqValid
    2/ nothing else!
    """
    pass


class EnumMod:
    _SOURCE: Any
    _EQ_VALID__CLS_DEF: Base_EqValid

    # ITEMS --------------
    ITEM1: Base_EqValid
    ITEM2: Base_EqValid
    ITEM3: Base_EqValid

    # DETECTED --------------
    _VALUE: Base_EqValid    # first validated EqValid
    _NAME: str              # original detected appropriate name

    def __eq__(self, other):
        # 1=same
        if isinstance(other, EnumMod):
            return self._NAME == other._NAME

        try:
            other_name = self.__class__(other)._NAME
            return self._NAME == other_name
        except:
            return False

    def __init__(self, source: Any) -> None | NoReturn:
        self._SOURCE = source
        self.init_finals()

    def init_finals(self) -> None | NoReturn:
        """
        GOAL
        ----
        1/ check source is acceptable value
        2/ apply _VALUE
        3/ detect relevant _NAME and apply
        """




























# =====================================================================================================================
class Type__:
    pass


class __EnumEqValid:
    """
    GOAL
    ----
    try create universal class with exact selecting items
    where no needs to create special logic for different cmpring
    """
    # ATTR1: Base_EqValid | Any
    # ATTR2: Base_EqValid | Any

    _SOURCE: Any
    _EQ_VALID__CLS_DEF: Base_EqValid

    def __init__(self, source: Any) -> None | NoReturn:
        for value in self.values():
            if source == self._EQ_VALID__CLS_DEF(value):
                return

        msg = f"{value=} not in {self.values()=}"
        raise Exc__Incompatible_Data(msg)

    @classmethod
    def cls__init_items(cls) -> None:
        for name in ():
            pass

    def find_item(self, item_draft: Any) -> Self:
        pass

    def names(self) -> Iterable[str]:
        pass

    def values(self) -> Iterable[Any]:
        pass

    def exists_in__names(self) -> bool:
        pass

    def exists_in__values(self) -> bool:
        pass















    # FIXME: cant create! - not work!!!
    # def __new__(cls, value_druft: Any):
    #     value_final = cls._value__get_original(value_druft)
    #     return cls(value_final)

    @classmethod
    def _value__get_original(cls, value_druft: Any) -> Any | NoValue:
        for item in cls:
            value_original = item.value
            if isinstance(value_druft, str):
                if value_druft.lower() == str(value_original).lower():
                    return value_original
            if value_druft == value_original:
                return value_original

        return NoValue

    # def __eq__(self, other) -> bool:
    #     if isinstance(other, self.__class__):
    #         return self.value == other.value        # or str(self.value).lower() == str(other.value).lower()    # NO!!!
    #
    #     else:
    #         for enum_i in self.__class__:
    #             if isinstance(other, str):
    #                 if str(enum_i.value).lower() == str(other).lower():
    #                     return True
    #             else:
    #                 try:
    #                     if enum_i.value == other or other == enum_i.value:
    #                         return True
    #                 except:
    #                     pass
    #     return False

    def __eq__(self, other) -> bool:
        # OBVIOUS -----------
        if isinstance(other, self.__class__):
            return self.value == other.value

        # CMP AVAILABLE -----
        cmp_available = False

        try:
            if other in self.__class__:
                cmp_available = True
        except:
            pass

        try:
            if other in self:
                cmp_available = True
        except:
            pass

        if cmp_available:
            return other == self.value or self == self.__class__(other)
        else:
            return False

    # TODO: add Contain classmeth???  cant understand! need metaclass!
    # @classmethod
    # def __contains__(cls, other) -> bool:
    #     if isinstance(other, cls):
    #         other = other.value
    #
    #     for enum_i in cls:
    #         if enum_i.value == other or str(enum_i.value).lower() == str(other).lower():
    #             return True
    #     else:
    #         return False


# =====================================================================================================================



import pytest

from base_aux.base_lambdas.m1_lambda import *
from base_aux.aux_attr.m2_annot3_gacls_keys_as_values import *


# =====================================================================================================================
class VictimCs(NestGaCls_AnnotNamesAsValuesCs):
    ATTR1: str
    ATTR2: str


class VictimIc(NestGaCls_AnnotNamesAsValuesIc):
    ATTR1: str
    ATTR2: str


Victim_VALUES = ("ATTR1", "ATTR2")


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="args, _EXPECTED",
    argvalues=[
        ("ATTR1", ["ATTR1", "ATTR1"]),
        ("attr1", [AttributeError, "ATTR1"]),

        ("ATTR2", ["ATTR2", "ATTR2"]),
        ("notExists", [AttributeError, AttributeError]),
        ("use spaces", [AttributeError, AttributeError]),
    ]
)
@pytest.mark.parametrize(argnames="Victim", argvalues=[VictimCs, VictimIc])
def test__ga(Victim, args, _EXPECTED):
    exp_index = 0
    if Victim is VictimCs:
        exp_index = 0
    elif Victim is VictimIc:
        exp_index = 1

    func_link = lambda value: getattr(Victim, value)
    Lambda(func_link, args).check_expected__assert(_EXPECTED[exp_index])


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="args, _EXPECTED",
    argvalues=[
        # INDEX ------
        (0, ["ATTR1", "ATTR1"]),
        (1, ["ATTR2", "ATTR2"]),
        (2, [IndexError, IndexError]),
        (-1, ["ATTR2", "ATTR2"]),
        (-2, ["ATTR1", "ATTR1"]),
        (-3, [IndexError, IndexError]),

        # ITEM ------
        ("ATTR1", ["ATTR1", "ATTR1"]),
        ("attr1", [AttributeError, "ATTR1"]),
        ("notExists", [AttributeError, AttributeError]),
        ("use spaces", [AttributeError, AttributeError]),
    ]
)
@pytest.mark.parametrize(argnames="Victim", argvalues=[VictimCs, VictimIc])
def test__gi(Victim, args, _EXPECTED):
    exp_index = 0
    if Victim is VictimCs:
        exp_index = 0
    elif Victim is VictimIc:
        exp_index = 1

    func_link = lambda value: Victim[value]
    Lambda(func_link, args).check_expected__assert(_EXPECTED[exp_index])


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="args, _EXPECTED",
    argvalues=[
        ("ATTR1", [True, True]),
        ("attr1", [False, True]),
        ("notExists", [False, False]),
        ("use spaces", [False, False]),
    ]
)
@pytest.mark.parametrize(argnames="Victim", argvalues=[VictimCs, VictimIc])
def test__in(Victim, args, _EXPECTED):
    exp_index = 0
    if Victim is VictimCs:
        exp_index = 0
    elif Victim is VictimIc:
        exp_index = 1

    func_link = lambda value: value in Victim
    Lambda(func_link, args).check_expected__assert(_EXPECTED[exp_index])


# =====================================================================================================================
def test__iter():
    assert tuple(VictimCs) == Victim_VALUES
    assert tuple(VictimIc) == Victim_VALUES


def test__len():
    assert len(VictimCs) == len(Victim_VALUES)
    assert len(VictimIc) == len(Victim_VALUES)

def test__str_repr():
    assert str(VictimCs) == str(Victim_VALUES) == "('ATTR1', 'ATTR2')"
    assert repr(VictimCs) == f"VictimCs{Victim_VALUES}" == "VictimCs('ATTR1', 'ATTR2')"

    assert str(VictimIc) == str(Victim_VALUES) == "('ATTR1', 'ATTR2')"
    assert repr(VictimIc) == f"VictimIc{Victim_VALUES}" == "VictimIc('ATTR1', 'ATTR2')"


# =====================================================================================================================



import pytest
import sys
import asyncio

from base_aux.base_lambdas.m1_lambda import *
from base_aux.base_types.m1_type_aux import *
from base_aux.base_values.m4_primitives import *


# =====================================================================================================================
class Test__1:
    @pytest.mark.parametrize(
        argnames="source, _EXPECTED",
        argvalues=[
            (pytest,      (False, False, False, False, False, True, )),
            (sys,      (False, False, False, False, False, True, )),

            (None,      (True, True, True, False, False, False, )),
            (True,      (True, True, True, True, False, False, )),
            (False,     (True, True, True, True, False, False, )),
            (0,         (False, True, True, True, False, False, )),
            (111,       (False, True, True, True, False, False, )),
            (111.222,   (False, True, True, True, False, False, )),
            ("str",     (False, True, True, True, False, False, )),
            (b"bytes",  (False, True, True, True, False, False, )),

            ((111, ),        (False, True, False, False, True, False, )),
            ([111, ],        (False, True, False, False, True, False, )),
            ({111, },        (False, True, False, False, True, False, )),
            ({111: 222, },   (False, True, False, False, True, False, )),

            (int,       (False, False, False, False, False, False, )),
            (int(1),    (False, True, True, True, False, False, )),
            (str,       (False, False, False, False, False, False, )),
            (str(1),    (False, True, True, True, False, False, )),

            (Exception,     (False, False, False, False, False, False, )),
            (Exception(),   (False, False, False, False, False, False, )),
            (ClsException,  (False, False, False, False, False, False, )),
            (ClsException(), (False, False, False, False,False, False, )),

            (Cls,       (False, False, False, False, False, False, )),
            (Cls(),     (False, False, False, False, False, False, )),
            (ClsInt,    (False, False, False, False, False, False, )),
            (ClsInt(),  (False, True, True, True, False, False, )),    # int() == 0!!!

            (FUNC,                      (False, False, False, False, False, False, )),
            (LAMBDA,                    (False, False, False, False, False, False, )),
            (ClsCallNone,               (False, False, False, False, False, False, )),
            (ClsCallNone(),             (False, False, False, False, False, False, )),
            (ClsCallNone()(),           (True, True, True, False, False, False, )),
            (ClsCall().meth,            (False, False, False, False, False, False, )),
            (ClsFullTypes.attrNone,     (True, True, True, False, False, False, )),
            (ClsFullTypes().attrNone,   (True, True, True, False, False, False, )),

            *[
                (
                    class_i,
                    (False, False, False, False, False, False, )
                ) for class_i in CLASSES__AS_FUNC
            ]
        ]
    )
    def test__check__bool_none(self, source, _EXPECTED):
        victim = TypeAux(source)

        Lambda(victim.check__bool_none).check_expected__assert(_EXPECTED[0])
        Lambda(victim.check__elementary).check_expected__assert(_EXPECTED[1])
        Lambda(victim.check__elementary_single).check_expected__assert(_EXPECTED[2])
        Lambda(victim.check__elementary_single_not_none).check_expected__assert(_EXPECTED[3])
        Lambda(victim.check__elementary_collection).check_expected__assert(_EXPECTED[4])
        Lambda(victim.check__module).check_expected__assert(_EXPECTED[5])





    @pytest.mark.parametrize(
        argnames="source, _EXPECTED",
        argvalues=[
            (None, False),
            (True, False),
            (False, False),
            (0, False),
            (111, False),
            (111.222, False),
            ("str", False),
            (b"bytes", False),

            ((111, ), True),
            ([111, ], True),
            ({111, }, True),
            ({111: 222, }, False),

            (int, False),
            (int(1), False),
            (str, False),
            (str(1), False),

            (Exception, False),
            (Exception(), False),
            (ClsException, False),
            (ClsException(), False),

            (Cls, False),
            (Cls(), False),
            (ClsInt, False),
            (ClsInt(), False),    # int() == 0!!!

            (FUNC, False),
            (LAMBDA, False),
            (ClsCallNone, False),
            (ClsCallNone(), False),
            (ClsCallNone()(), False),
            (ClsCall.meth, False),
            (ClsCall().meth, False),
            (ClsFullTypes.attrNone, False),
            (ClsFullTypes().attrNone, False),

            *[(class_i, False) for class_i in CLASSES__AS_FUNC]
        ]
    )
    def test__check__elementary_collection_not_dict(self, source, _EXPECTED):
        func_link = TypeAux(source).check__elementary_collection_not_dict
        Lambda(func_link).check_expected__assert(_EXPECTED)

    # -----------------------------------------------------------------------------------------------------------------
    @pytest.mark.parametrize(
        argnames="source, args, _EXPECTED",
        argvalues=[
            ("str", (True, True), True),
            ("str", (True, False), False),

            (b"bytes", (True, True), True),
            (b"bytes", (True, False), False),

            # -----------------------
            (None, (), False),
            (True, (), False),
            (False, (), False),
            (0, (), False),
            (111, (), False),
            (111.222, (), False),
            ("str", (), True),
            (b"bytes", (), True),

            ((111, ), (), True),
            ([111, ], (), True),
            ({111, }, (), True),
            ({111: 222, }, (), True),
            ({111: 222, }, (True, True), True),
            ({111: 222, }, (False, True), False),

            (int, (), False),
            (int(1), (), False),
            (str, (), True),        # not clear!!!
            (str(1), (), True),

            (Exception, (), False),
            (Exception(), (), False),
            (ClsException, (), False),
            (ClsException(), (), False),

            (Cls, (), False),
            (Cls(), (), False),
            (ClsInt, (), False),
            (ClsInt(), (), False),

            (FUNC, (), False),
            (LAMBDA, (), False),
            (ClsCallNone, (), False),
            (ClsCallNone(),(),  False),
            (ClsCallNone()(), (), False),
            (ClsCall.meth, (), False),
            (ClsCall().meth, (), False),
            (ClsFullTypes.attrNone, (), False),
            (ClsFullTypes().attrNone, (), False),

            # *[(class_i, False) for class_i in CLASSES__AS_FUNC]
        ]
    )
    def test__iterable(self, source, args, _EXPECTED):
        func_link = TypeAux(source).check__iterable
        Lambda(func_link, *args).check_expected__assert(_EXPECTED)

    @pytest.mark.parametrize(
        argnames="source, _EXPECTED",
        argvalues=[
            (None, False),
            (True, False),
            (False, False),
            (0, False),
            (111, False),
            (111.222, False),
            ("str", False),
            (b"bytes", False),

            (((111, ),), True),
            (([111, ],), True),
            (({111, },), True),
            (({111: 222, },), True),

            (int, False),
            (int(1), False),
            (str, True),        # not clear!!!
            (str(1), False),

            (Exception, False),
            (Exception(), False),
            (ClsException, False),
            (ClsException(), False),

            (Cls, False),
            (Cls(), False),
            (ClsInt, False),
            (ClsInt(), False),

            (FUNC, False),
            (LAMBDA, False),
            (ClsCallNone, False),
            (ClsCallNone(), False),
            (ClsCallNone()(), False),
            (ClsCall.meth, False),
            (ClsCall().meth, False),
            (ClsFullTypes.attrNone, False),
            (ClsFullTypes().attrNone, False),

            # *[(class_i, False) for class_i in CLASSES__AS_FUNC]
        ]
    )
    def test__iterable_not_str(self, source, _EXPECTED):
        func_link = TypeAux(source).check__iterable_not_str
        Lambda(func_link).check_expected__assert(_EXPECTED)

    # CALLABLE --------------------------------------------------------------------------------------------------------
    @pytest.mark.parametrize(
        argnames="source, _EXPECTED",
        argvalues=[
            (None, False),
            (True, False),
            (False, False),
            (0, False),
            (111, False),
            (111.222, False),
            ("str", False),
            (b"bytes", False),

            (((111, ),), False),
            (([111, ],), False),
            (({111, },), False),
            (({111: 222, },), False),

            (int, True),
            (int(1), False),
            (str, True),
            (str(1), False),

            (Exception, False),
            (Exception(), False),
            (ClsException, False),
            (ClsException(), False),

            (Cls, False),
            (Cls(), False),
            (ClsInt, True),
            (ClsInt(), False),    # int() == 0!!!

            (FUNC, True),
            (LAMBDA, True),
            (ClsCallNone, False),
            (ClsCallNone(), True),
            (ClsCallNone()(), False),
            (ClsCall.meth, True),
            (ClsCall().meth, True),
            (ClsFullTypes.attrNone, False),
            (ClsFullTypes().attrNone, False),

            *[(class_i, True) for class_i in CLASSES__AS_FUNC]
        ]
    )
    def test__check__callable_func_meth_inst(self, source, _EXPECTED):
        func_link = TypeAux(source).check__callable_func_meth_inst
        Lambda(func_link).check_expected__assert(_EXPECTED)

    @pytest.mark.parametrize(
        argnames="source, _EXPECTED",
        argvalues=[
            (None, False),
            (True, False),
            (False, False),
            (0, False),
            (111, False),
            (111.222, False),
            ("str", False),
            (b"bytes", False),

            (((111, ),), False),
            (([111, ],), False),
            (({111, },), False),
            (({111: 222, },), False),

            (int, True),
            (int(1), False),
            (str, True),
            (str(1), False),

            (Exception, False),
            (Exception(), False),
            (ClsException, False),
            (ClsException(), False),

            (Cls, False),
            (Cls(), False),
            (ClsInt, True),
            (ClsInt(), False),    # int() == 0!!!

            (FUNC, True),
            (LAMBDA, True),
            (ClsCallNone, False),
            (ClsCallNone(), False),
            (ClsCallNone()(), False),
            (ClsCall.meth, True),
            (ClsCall().meth, True),
            (ClsFullTypes.attrNone, False),
            (ClsFullTypes().attrNone, False),

            *[(class_i, True) for class_i in CLASSES__AS_FUNC]
        ]
    )
    def test__check__callable_func_meth(self, source, _EXPECTED):
        func_link = TypeAux(source).check__callable_func_meth
        Lambda(func_link).check_expected__assert(_EXPECTED)

    @pytest.mark.parametrize(
        argnames="source, _EXPECTED",
        argvalues=[
            (None, False),
            (True, False),
            (False, False),
            (0, False),
            (111, False),
            (111.222, False),
            ("str", False),
            (b"bytes", False),

            (((111, ),), False),
            (([111, ],), False),
            (({111, },), False),
            (({111: 222, },), False),

            (int, True),
            (int(1), False),
            (str, True),
            (str(1), False),

            (Exception, False),
            (Exception(), False),
            (ClsException, False),
            (ClsException(), False),

            (Cls, False),
            (Cls(), False),
            (ClsInt, True),
            (ClsInt(), False),    # int() == 0!!!

            (FUNC, True),
            (LAMBDA, True),
            (ClsCallNone, False),
            (ClsCallNone(), False),
            (ClsCallNone()(), False),
            (ClsCall.meth, True),
            (ClsCall().meth, False),
            (ClsFullTypes.attrNone, False),
            (ClsFullTypes().attrNone, False),

            *[(class_i, True) for class_i in CLASSES__AS_FUNC]
        ]
    )
    def test__check__callable_func(self, source, _EXPECTED):
        func_link = TypeAux(source).check__callable_func
        Lambda(func_link).check_expected__assert(_EXPECTED)

    @pytest.mark.parametrize(
        argnames="source, _EXPECTED",
        argvalues=[
            (None, False),
            (True, False),
            (False, False),
            (0, False),
            (111, False),
            (111.222, False),
            ("str", False),
            (b"bytes", False),

            (((111, ),), False),
            (([111, ],), False),
            (({111, },), False),
            (({111: 222, },), False),

            (int, False),
            (int(1), False),
            (str, False),
            (str(1), False),

            (Exception, False),
            (Exception(), False),
            (ClsException, False),
            (ClsException(), False),

            (Cls, False),
            (Cls(), False),
            (ClsInt, False),
            (ClsInt(), False),    # int() == 0!!!

            (FUNC, False),
            (LAMBDA, False),
            (ClsCallNone, False),
            (ClsCallNone(), False),
            (ClsCallNone()(), False),
            (ClsCall.meth, False),
            (ClsCall().meth, True),
            (ClsFullTypes.attrNone, False),
            (ClsFullTypes().attrNone, False),

            *[(class_i, False) for class_i in CLASSES__AS_FUNC]
        ]
    )
    def test__check__callable_meth(self, source, _EXPECTED):
        func_link = TypeAux(source).check__callable_meth
        Lambda(func_link).check_expected__assert(_EXPECTED)

    @pytest.mark.parametrize(
        argnames="source, _EXPECTED",
        argvalues=[
            (None, False),
            (True, False),
            (False, False),
            (0, False),
            (111, False),
            (111.222, False),
            ("str", False),
            (b"bytes", False),

            (((111, ),), False),
            (([111, ],), False),
            (({111, },), False),
            (({111: 222, },), False),

            (int, False),
            (int(1), False),
            (str, False),
            (str(1), False),

            (Exception, False),
            (Exception(), False),
            (ClsException, False),
            (ClsException(), False),

            (Cls, False),
            (Cls(), False),
            (ClsInt, False),
            (ClsInt(), False),    # int() == 0!!!

            (FUNC, False),
            (LAMBDA, False),
            (ClsCallNone, False),
            (ClsCallNone(), True),
            (ClsCallNone()(), False),
            (ClsCall.meth, False),
            (ClsCall().meth, False),
            (ClsFullTypes.attrNone, False),
            (ClsFullTypes().attrNone, False),

            *[(class_i, False) for class_i in CLASSES__AS_FUNC]
        ]
    )
    def test__check__callable_inst(self, source, _EXPECTED):
        func_link = TypeAux(source).check__callable_inst
        Lambda(func_link).check_expected__assert(_EXPECTED)

    @pytest.mark.parametrize(
        argnames="source, _EXPECTED",
        argvalues=[
            (None, False),
            (True, False),
            (False, False),
            (0, False),
            (111, False),
            (111.222, False),
            ("str", False),
            (b"bytes", False),

            (((111, ),), False),
            (([111, ],), False),
            (({111, },), False),
            (({111: 222, },), False),

            (int, True),
            (int(1), False),
            (str, True),
            (str(1), False),

            (Exception, False),
            (Exception(), False),
            (ClsException, False),
            (ClsException(), False),

            (Cls, False),
            (Cls(), False),
            (ClsInt, True),
            (ClsInt(), False),    # int() == 0!!!

            (FUNC, False),
            (LAMBDA, False),
            (ClsCallNone, False),
            (ClsCallNone(), False),
            (ClsCallNone()(), False),
            (ClsCall.meth, False),
            (ClsCall().meth, False),
            (ClsFullTypes.attrNone, False),
            (ClsFullTypes().attrNone, False),

            *[(class_i, True) for class_i in CLASSES__AS_FUNC]
        ]
    )
    def test__check__callable_cls_as_func_builtin(self, source, _EXPECTED):
        func_link = TypeAux(source).check__callable_cls_as_func_builtin
        Lambda(func_link).check_expected__assert(_EXPECTED)

    # -----------------------------------------------------------------------------------------------------------------
    @pytest.mark.parametrize(
        argnames="source, _EXPECTED",
        argvalues=[
            (None, False),
            (True, False),
            (False, False),
            (0, False),
            (111, False),
            (111.222, False),
            ("str", False),
            (b"bytes", False),

            (((111, ),), False),
            (([111, ],), False),
            (({111, },), False),
            (({111: 222, },), False),

            (int, True),
            (int(1), False),
            (str, True),
            (str(1), False),

            (Exception, True),
            (Exception(), False),
            (ClsException, True),
            (ClsException(), False),

            (Cls, True),
            (Cls(), False),
            (ClsInt, True),
            (ClsInt(), False),    # int() == 0!!!

            (FUNC, False),
            (LAMBDA, False),
            (ClsCallNone, True),
            (ClsCallNone(), False),
            (ClsCallNone()(), False),
            (ClsCall.meth, False),
            (ClsCall().meth, False),
            (ClsFullTypes.attrNone, False),
            (ClsFullTypes().attrNone, False),

            *[(class_i, True) for class_i in CLASSES__AS_FUNC]
        ]
    )
    def test__check__class(self, source, _EXPECTED):
        func_link = TypeAux(source).check__class
        Lambda(func_link).check_expected__assert(_EXPECTED)

    @pytest.mark.parametrize(
        argnames="source, _EXPECTED",
        argvalues=[
            (None, True),
            (True, True),
            (False, True),
            (0, True),
            (111, True),
            (111.222, True),
            ("str", True),
            (b"bytes", True),

            (((111, ),), True),
            (([111, ],), True),
            (({111, },), True),
            (({111: 222, },), True),

            (int, False),
            (int(1), True),
            (str, False),
            (str(1), True),

            (Exception, False),
            (Exception(), True),
            (ClsException, False),
            (ClsException(), True),

            (Cls, False),
            (Cls(), True),
            (ClsInt, False),
            (ClsInt(), True),    # int() == 0!!!

            (FUNC, False),
            (LAMBDA, False),
            (ClsCallNone, False),
            (ClsCallNone(), True),
            (ClsCallNone()(), True),
            (ClsCall.meth, False),
            (ClsCall().meth, False),
            (ClsFullTypes.attrNone, True),
            (ClsFullTypes().attrNone, True),

            *[(class_i, False) for class_i in CLASSES__AS_FUNC]
        ]
    )
    def test__check__instance(self, source, _EXPECTED):
        func_link = TypeAux(source).check__instance
        Lambda(func_link).check_expected__assert(_EXPECTED)

    @pytest.mark.parametrize(
        argnames="source, _EXPECTED",
        argvalues=[
            (None, False),
            (True, False),
            (False, False),
            (0, False),
            (111, False),
            (111.222, False),
            ("str", False),
            (b"bytes", False),

            (((111, ),), False),
            (([111, ],), False),
            (({111, },), False),
            (({111: 222, },), False),

            (int, False),
            (int(1), False),
            (str, False),
            (str(1), False),

            (Exception, False),
            (Exception(), True),
            (ClsException, False),
            (ClsException(), True),

            (Cls, False),
            (Cls(), True),
            (ClsInt, False),
            (ClsInt(), False),    # int() == 0!!!

            (FUNC, False),
            (LAMBDA, False),
            (ClsCallNone, False),
            (ClsCallNone(), True),
            (ClsCallNone()(), False),
            (ClsCall.meth, False),
            (ClsCall().meth, False),
            (ClsFullTypes.attrNone, False),
            (ClsFullTypes().attrNone, False),

            *[(class_i, False) for class_i in CLASSES__AS_FUNC]
        ]
    )
    def test__check__instance_not_elementary(self, source, _EXPECTED):
        func_link = TypeAux(source).check__instance_not_elementary
        Lambda(func_link).check_expected__assert(_EXPECTED)

    # -----------------------------------------------------------------------------------------------------------------
    @pytest.mark.parametrize(
        argnames="source, _EXPECTED",
        argvalues=[
            (None, False),
            (True, False),
            (False, False),
            (0, False),
            (111, False),
            (111.222, False),
            ("str", False),
            (b"bytes", False),

            (((111, ),), False),
            (([111, ],), False),
            (({111, },), False),
            (({111: 222, },), False),

            (int, False),
            (int(1), False),
            (str, False),
            (str(1), False),

            (Exception, True),
            (Exception(), True),
            (ClsException, True),
            (ClsException(), True),

            (Cls, False),
            (Cls(), False),
            (ClsInt, False),
            (ClsInt(), False),    # int() == 0!!!

            (FUNC, False),
            (LAMBDA, False),
            (ClsCallNone, False),
            (ClsCallNone(), False),
            (ClsCallNone()(), False),
            (ClsCall.meth, False),
            (ClsCall().meth, False),
            (ClsFullTypes.attrNone, False),
            (ClsFullTypes().attrNone, False),

            *[(class_i, False) for class_i in CLASSES__AS_FUNC]
        ]
    )
    def test__check__exception(self, source, _EXPECTED):
        func_link = TypeAux(source).check__exception
        Lambda(func_link).check_expected__assert(_EXPECTED)

    # =================================================================================================================
    @pytest.mark.parametrize(
        argnames="index",
        argvalues=[
            0,
            1,
    ])
    @pytest.mark.parametrize(
        argnames="source, parent, _EXPECTED",
        argvalues=[
            ("str", ("str", str), [True, True]),
            ("str", (str, "str"), [True, True]),

            ("str", "str", [True, False]),
            ("str", str, [True, True]),
            (str, "str", [True, False]),
            (str, str, [True, True]),

            (int, str, [False, False]),
            (int, "str", [False, False]),

            (111, 111, [True, False]),
            (int, 111, [True, False]),
            (111, int, [True, True]),
            (int, int, [True, True]),

            (Exception, Exception, [True, True]),
            (Exception(), Exception, [True, True]),
            (Exception, Exception(), [True, False]),
            (Exception(), Exception(), [True, False]),

            (ClsException, Exception, [True, True]),
            (ClsException(), Exception, [True, True]),
            (ClsException, Exception(), [True, False]),
            (ClsException(), Exception(), [True, False]),

            (Exception, ClsException, [False, False]),      # REMEMBER! not clear!
            (Exception(), ClsException, [False, False]),    # REMEMBER! not clear!
            (Exception, ClsException(), [False, False]),    # REMEMBER! not clear!
            (Exception(), ClsException(), [False, False]),  # REMEMBER! not clear!

            (Cls, Cls, [True, True]),
            (Cls, Cls(), [True, False]),
            (Cls(), Cls, [True, True]),
            (Cls(), Cls(), [True, False]),

            (FUNC, Cls, [False, False]),
            (FUNC, Cls(), [False, False]),
        ]
    )
    def test__check__nested__by_cls_or_inst(self, source, parent, _EXPECTED, index):
        if index == 0:
            func_link = TypeAux(source).check__subclassed_or_isinst__from_cls_or_inst
        elif index == 1:
            func_link = TypeAux(source).check__subclassed_or_isinst__from_cls
        else:
            raise Exception(f"incorrect {index=}")

        if isinstance(parent, (tuple, list)):
            Lambda(func_link, *parent).check_expected__assert(_EXPECTED[index])
        else:
            Lambda(func_link, parent).check_expected__assert(_EXPECTED[index])

    # =================================================================================================================
    @pytest.mark.parametrize(
        argnames="source, _EXPECTED",
        argvalues=[
            (None, (False, False)),
            (asyncio.sleep, (True, False)),
            (asyncio.sleep(1), (False, True)),
        ]
    )
    def test__check__aio(self, source, _EXPECTED):
        func_link = TypeAux(source).check__coro_func
        Lambda(func_link).check_expected__assert(_EXPECTED[0])

        func_link = TypeAux(source).check__coro
        Lambda(func_link).check_expected__assert(_EXPECTED[1])


# =====================================================================================================================



import pytest

from base_aux.base_lambdas.m1_lambda import *
from base_aux.classes.m2_middle_group import *


# =====================================================================================================================
class Victim(ClsMiddleGroup):
    pass


class Victim1(ClsMiddleGroup):
    MIDDLE_GROUP__NAME = "name1"


class Victim2(ClsMiddleGroup):
    MIDDLE_GROUP__NAME = "name2"


class VictimAttr1(ClsMiddleGroup):
    attr = 1


class VictimAttr1Cmp(VictimAttr1):
    MIDDLE_GROUP__CMP_ATTR = "attr"


class VictimAttr1CmpAttr2(VictimAttr1Cmp):
    attr = 2


class VictimAttr1CmpAttr2_TryBreak(VictimAttr1CmpAttr2):
    MIDDLE_GROUP__CMP_ATTR = "tryBreak"


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="obj, other, _EXPECTED",
    argvalues=[
        (Victim, True, None),
        (Victim(), True, None),
        (Victim, 1, None),
        (Victim(), 1, None),
        (Victim, bool, None),
        (Victim(), bool, None),

        (Victim, Victim, False),
        (Victim, Victim(), False),
        (Victim, Victim1, False),
        (Victim, Victim1(), False),

        # VictimAttr1
        (Victim, VictimAttr1, False),
        (Victim, VictimAttr1Cmp, False),

        (VictimAttr1, VictimAttr1CmpAttr2, False),

        (VictimAttr1Cmp, VictimAttr1, True),
        (VictimAttr1Cmp, VictimAttr1CmpAttr2, False),

        (VictimAttr1CmpAttr2_TryBreak, VictimAttr1CmpAttr2, True),
    ]
)
def test__victim__check_equal__cls(obj, other, _EXPECTED):
    Lambda(obj.middle_group__check_equal__cls, other).check_expected__assert(_EXPECTED)


# =====================================================================================================================
